classdef simpleModelDC < simpleModel
    % SIMPLEMODELCSVM This class inherits from simpleModel. It contains
    % particular methods to solve binary-C-SVM.
    % This class support C-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 9/01/2014
    % Last Modified 9/01/2014 by G.L.
    
    properties
        prevalpha;
        Gneg;
    end
    
    
    methods
        % contructor
        function SM = simpleModelDC(SD,C,Il,Iu,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelCSVM object
            
            assert(nargin==2 || nargin==5,'simpleModelCSVM:wrongInput','Requires 2 or 5 input arguments');
            if nargin<2
                error('not enough input');
            else
                
                SD.getKernelValues(1:SD.n); % pre-compute kernel
                [V D] = eig(0.5*SD.kernelCache(1:SD.n,1:SD.n)'+0.5*SD.kernelCache(1:SD.n,1:SD.n));
                negs = find(diag(D)<0);
                poss = find(diag(D)>=0);
                SD.kernelCache(1:SD.n,1:SD.n) = V(:,poss)*D(poss,poss)*V(:,poss)';
                SM.Gneg = -V(:,negs)*D(negs,negs)*V(:,negs)';
                
                SM.SD = SD;
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                if nargin==5
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [find(SD.trainlab==-1,1),find(SD.trainlab==1,1)];
                    SM.Iu = [];
                    SM.Il = 1:SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(2,1);
                end
                SM.labs = sort(unique(SD.trainlab));
                if length(C)==length(SD.trainlab)
                    SM.Slacks = C;
                elseif length(C)==length(SM.labs)
                    for i=1:length(SM.labs)
                        SM.Slacks(SD.trainlab==SM.labs(i)) = C(i);
                    end
                else
                    SM.Slacks = C(1)*ones(size(SD.trainlab));
                    
                end
                SM.activated = 0;
                SM.converged = 0;
            end
        end
        
        
        function train(obj)
            obj.prevalpha = zeros(obj.SD.n,1);
            iter2 = 0;
            while iter2<30
                obj.converged = 0;
                iter = 0;
                while obj.converged==0 && iter<500*length(obj.SD.n)
                    iter = iter+1;
                    obj.solveLinearSystem();
                    %obj.print2D();
                    %keyboard
                    obj.activateBoxConstraints();
                    if (obj.activated==0)
                        obj.relaxBoxConstraints();
                    end
                end
                if obj.converged==0
                    disp('No convergence');
                    obj.solveLinearSystem();
                    obj.converged = 1;
                end
                %obj.print2D();
                obj.prevalpha = zeros(obj.SD.n,1);
                obj.prevalpha([obj.Iw,obj.Iu]) = [obj.alpha;obj.Slacks(obj.Iu)];
                iter2 = iter2+1;
                %nul = obj.evaluateSolution(1:obj.SD.n)+(obj.prevalpha([obj.Iw,obj.Iu])'*obj.Gneg([obj.Iw,obj.Iu],1:obj.SD.n))';
                %err = length(find(nul<0))
            end
            obj.finish();
            
        end
        
        
        
        function set.prevalpha(obj,val)
            obj.prevalpha = val;
        end
        
        function val = get.prevalpha(obj)
            val = obj.prevalpha;
        end
        
        function set.Gneg(obj,val)
            obj.Gneg = val;
        end
        
        function val = get.Gneg(obj)
            val = obj.Gneg;
        end
        
        
        function y = evaluateSolution(obj,candidates)
            indd = obj.SD.getKernelValues([obj.Iw,obj.Iu],candidates);
            [val,inds]= sort(indd);
            temp = [obj.alpha;obj.Slacks(obj.Iu)];
            y = obj.SD.kernelCache(candidates,val)*temp(inds)+obj.SD.trainlab(candidates)*obj.b-(obj.prevalpha([obj.Iw,obj.Iu])'*obj.Gneg([obj.Iw,obj.Iu],candidates))';
        end
        
        function solveLinearSystem(obj)
            %indd = obj.SD.getKernelValues([obj.Iw,obj.Iu],obj.Iw);
            ce = ones(1,length(obj.Iw))+obj.prevalpha([obj.Iw,obj.Iu])'*obj.Gneg([obj.Iw,obj.Iu],obj.Iw);
            be = 0;
            indd = obj.SD.getKernelValues(obj.Iw,obj.Iu);
            if ~isempty(obj.Iu)
                ce = ce - (obj.Slacks(obj.Iu))'*obj.SD.kernelCache(obj.Iu,indd);
                be = be - obj.Slacks(obj.Iu)'*obj.SD.trainlab(obj.Iu);
            end
            if obj.optimType==1
                G = chol(obj.SD.kernelCache(obj.Iw,indd)+obj.kktgap*eye(length(obj.Iw)));
                M = obj.SD.trainlab(obj.Iw)'*(G\(G'\obj.SD.trainlab(obj.Iw)));
                obj.b = M\(obj.SD.trainlab(obj.Iw)'*(G\(G'\ce'))-be);
                obj.alpha = G\(G'\(ce'-obj.SD.trainlab(obj.Iw)*obj.b));
            else
                error('not implemented yet');
            end
        end
        
        
        function reinit(obj)
            obj.Iw = [find(obj.SD.trainlab==-1,1),find(obj.SD.trainlab==1,1)];
            obj.Iu = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            
            if length(obj.Slacks)~=length(obj.SD.trainlab)
                if length(obj.C)==length(obj.SD.trainlab)
                    obj.Slacks = obj.C;
                elseif length(obj.C)==length(obj.labs)
                    for i=1:length(obj.labs)
                        obj.Slacks(obj.SD.trainlab==obj.labs(i)) = obj.C(i);
                    end
                else
                    obj.Slacks = obj.C(1)*ones(size(obj.SD.trainlab));
                    
                end
            end
            obj.activated = 0;
            obj.converged = 0;
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
            %f = kernelEval(obj,xtest);
            f = kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ...
                    abs(obj.SD.trainlab([obj.Iw,obj.Iu])), xtest, ones(size(xtest,2),1))';
            if obj.converged==0
                ypred = ([obj.alpha;obj.Slacks(obj.Iu)].*obj.SD.trainlab([obj.Iw,obj.Iu]))'*f+obj.b;
            else
                ypred = obj.alpha'*f+obj.b;
            end
            ypredmat=reshape(ypred,na,nb);
            
            hold off;
            plot(obj.SD.trainvec(1,:),obj.SD.trainvec(2,:),'k+');
            hold on;
            plot(obj.SV(1,:),obj.SV(2,:),'om','MarkerSize',10);        
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ 0 0],'g');
            set(h,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ 1 1],'y');
            set(h,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ -1 -1],'m');
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        
        
        function print2DMulti(obj,SMt)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            assert(nargin==2);
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            if length(SMt)==length(unique(obj.SD.trainlab));
                ypred = obj.test1vsAll(SMt,xtest);
            else
                ypred = obj.test1vs1(SMt,xtest);
            end
            ypredmat=reshape(ypred,na,nb);
            %labs = unique(obj.SD.trainlab);
            hold on;
            [c,h] = contour(xtesta1,xtesta2,ypredmat);
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
    end
    
end
