classdef simpleDataLS < simpleData
    % SIMPLEDATALS This class contains all data that are used in simpleModel
    % solvers. It stores initial data , computes
    % the kernel matrix on-demand for large-scale data. No pre-computed
    % kernel here.
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 16/02/2011
    % Last Modified 16/02/2011 by G.L.
    
    properties
        trainvec; % stored as column individuals
        trainlab; % stored as column
        kernelType = 'rbf';
        kernelParam = 1;
        kernelCache; % computed on the fly 
        n;
    end
    
    properties (SetAccess = private)
        indx; % cached rows
        relIndx; % position of cached rows
        premult;
    end
    
    
    methods
        % contructor
        function SD = simpleData(tt,tl)
            assert(nargin==2,'simpleDataLS:wrongInput','SimpleData needs training points or pre-computed kernel matrix and associated labels : two arguments');
            assert(size(tl,1)==size(tt,2) && size(tl,2)==1,'simpleDataLS:wrongInput','Labels should be in a column vector, training point stored as columns');
            
            SD.trainvec = tt;
            SD.trainlab = tl;
            SD.relIndx = zeros(1,size(tt,2));
            SD.indx = zeros(1,size(tt,2));
            SD.n = size(tt,2);
            SD.premult = tt.*tt;
            SD.kernelCache = zeros(100,100);
        end
        
       
        
        function K = getKernelValues(obj,indx,indy)
            if nargin==2
                %a = indx(~(ismember(indx,obj.indx)));
                a = find(obj.indx(indx)==0,1);
                if ~isempty(a)
                    obj.computeKernel(a);
                    obj.updateCache();
                end
                K = obj.relIndx(indx);
            elseif nargin==3
                K = kernelLab(obj,indx,indy);
            end
        end
        
        function obj = computeKernel(obj,ind)
            relInd = find(obj.relInd==0,1);
            obj.indx(ind) = relInd;
            obj.relInd(obj.indx(ind)) = ind;
            obj.kernelCache(obj.relInd>0,relInd) = kernelLab(obj,obj.indx(obj.relInd>0),ind);
        end
        
        
        function indd = getDocKernelValues(obj,indx,indy,Q,lambda)
            %a = setdiff(indx, obj.indx);
            a = indx(~(ismember(indx,obj.indx)));
            if ~isempty(a)
                obj.computeDocKernel(a,Q,lambda);
                obj.updateCache();
            end
            %K = obj.kernelCache(indy,obj.relIndx(indx));
            indd = obj.relIndx(indx);
        end
        
        function obj = computeDocKernel(obj,ind,Q,lambda)
            relInd = length(obj.indx)+1:length(obj.indx)+length(ind);
            obj.kernelDocLab(ind,Q,lambda,relInd);
            obj.relIndx(ind) = relInd;
            obj.indx = [obj.indx,ind];
        end
        
        function f = kernelLab(obj,ind,indy)
            if strcmp(obj.kernelType,'rbf')
                if nargin==2
                    dot_b = sum(obj.premult(:,ind),1);
                    dot_a = sum(obj.premult,1)';
                    unitvec = ones(obj.n,1);
                    f =   obj.trainvec'*obj.trainvec(:,ind);
                    for i=1:length(ind)
                        a = obj.kernelParam(1) * ...
                            (2 * f(:,i) - dot_a - dot_b(i) * unitvec);
                        f(:,i) = obj.trainlab(ind(i)) * (exp(a) .* obj.trainlab);
                    end
                elseif nargin==3
                    dot_b = sum(obj.premult(:,ind),1);
                    dot_a = sum(obj.premult(:,indy),1)';
                    unitvec = ones(length(indy),1);
                    f =   obj.trainvec(:,indy)'*obj.trainvec(:,ind);
                    for i=1:length(ind)
                        a = obj.kernelParam(1) * ...
                            (2 * f(:,i) - dot_a - dot_b(i) * unitvec);
                        f(:,i) = obj.trainlab(ind(i)) * (exp(a) .* obj.trainlab(indy));
                    end
                end
            end
        end
        
        function obj = kernelDocLab(obj,ind,Q,lambda,relInd)
            if strcmp(obj.kernelType,'rbf')
                dot_b = sum(obj.premult(:,ind),1);
                dot_a = sum(obj.premult,1)';
                unitvec = ones(obj.n,1);
                A = lambda/(1+Q*lambda);
                obj.kernelCache(:,relInd) =   obj.trainvec'*obj.trainvec(:,ind);
                if lambda==0
                    for i=1:length(ind)
                        obj.kernelCache(:,relInd(i)) =(obj.trainlab(ind(i))==obj.trainlab) .* (exp(obj.kernelParam(1) * ...
                            (2 * obj.kernelCache(:,relInd(i)) - dot_a - dot_b(i) * unitvec)));
                    end
                else
                    B = 1/lambda + (Q-1);
                    for i=1:length(ind)
                        obj.kernelCache(:,relInd(i)) =(1-(B-1)*(obj.trainlab(ind(i))==obj.trainlab)) .* (-A*exp(obj.kernelParam(1) * ...
                            (2 * obj.kernelCache(:,relInd(i)) - dot_a - dot_b(i) * unitvec)));
                    end
                end
            end
        end
        
        
        
        function updateCache(obj)
            if length(obj.indx)+10>size(obj.kernelCache,2)
                obj.kernelCache = [obj.kernelCache,zeros(obj.n,50)];
            end
        end
        
        
        function clear(obj)
            obj.indx = [];
            obj.relIndx = obj.relIndx*0;
        end
        
    end
    
end

