classdef simpleKMKLModel < simpleMKLModel
    % SIMPLEMKLMODELCSVM This class inherits from simpleModel. It contains
    % particular methods to solve Support Vector Data Description
    % This class support SVDD solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 29/11/2013
    % Last Modified 29/11/2013 by G.L.
    
    properties
        nbEig;
    end
    
    methods
        % contructor
        function SM = simpleKMKLModel(SD,C,nbEig,Il,Iu,Iw)
            % simpleModelSVDD : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all
            %           points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelSVDD object
            
            assert(nargin==3 || nargin==6 ,'simpleKMKLModel:wrongInput','Requires 3 or 6 input arguments');
            if nargin<2
                error('not enough input');
            else
                SD.getKernelValues(1:SD.n); % pre-compute kernel
                SM.SD = SD;
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                if nargin>=5
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [find(SD.trainlab==-1,1),find(SD.trainlab==1,1)];
                    SM.Iu = [];
                    SM.Il = 1:SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(2,1);
                end
                SM.dm = ones(1,SD.nbKernels)./SD.nbKernels;
                
                SM.labs = sort(unique(SD.trainlab));
                if length(C)==length(SD.trainlab)
                    SM.Slacks = C;
                elseif length(C)==length(SM.labs)
                    for i=1:length(SM.labs)
                        SM.Slacks(SD.trainlab==SM.labs(i)) = C(i);
                    end
                else
                    SM.Slacks = C(1)*ones(size(SD.trainlab));
                    
                end
                SM.activated = 0;
                SM.converged = 0;
                SM.nbEig = nbEig;
            end
        end
        
        
        
        function val = get.nbEig(obj)
            val = obj.nbEig;
        end
        
        
        function set.nbEig(obj,val)
            obj.nbEig = val;
        end
        
        
        function [Jcross] = updateGrads(obj,SM,J,goldensearch_deltmax)
            [~,mu] = max(obj.dm);
            for m=1:length(obj.dm)
                dJ(m) = -0.5*SM.alpha'*(obj.SD.kernelCachePart(m,[SM.Iw,SM.Iu]))*SM.alpha;
            end
            NormGrad = dJ*dJ';
            dJ=dJ/sqrt(NormGrad);
            
            dJ = dJ - dJ(mu);
            D = -dJ.*( (obj.dm>0) | (dJ<0));
            D(mu) = -sum(D);
            
            gamma_min = 0;
            Jcross = J-1;
            
            ind = find(D<0);
            gamma = min(-(obj.dm(ind))./D(ind));
            deltmax = gamma;
            if isempty(gamma) || gamma==0
                obj.dm = obj.dm ;
                return
            end
            if gamma > 1
                gamma=1;
            end;
            
            while (J>Jcross)
                dcross = obj.dm+gamma*D;
                K = obj.SD.computeKernelCombination(dcross,1:obj.SD.n);
                SD = simpleData(K,obj.SD.trainlab);
                SD.kernelType='precomputed';
                SM.clear;
                SM = simpleModelENPSVM(SD,obj.C,obj.nbEig,SM.Il, SM.Iu, SM.Iw);
                SM.train;
                Jcross = (-0.5*(SM.alpha)'*K([SM.Iw,SM.Iu],[SM.Iw,SM.Iu])*SM.alpha + sum(SM.alpha.*obj.SD.trainlab));
                if (Jcross<J)
                    J = Jcross;
                    obj.dm = dcross;
                    D = D.*(dcross>10^-8 | D>0);
                    D(mu) =  - sum(D([1:mu-1,mu+1:end]));
                    ind = find(D<0);
                    obj.Iu = SM.Iu;
                    obj.Iw = SM.Iw;
                    obj.Il = SM.Il;
                    obj.alpha = SM.alpha;
                    obj.b = SM.b;
                    
                    if ~isempty(ind)
                        gamma = min(-(obj.dm(ind))./D(ind));
                        deltmax = gamma;
                        Jcross = 0;
                    else
                        gamma = 0;
                        deltmax = 0;
                    end
                    if gamma > 0.1
                        gamma=0.1;
                    end;
                end
                
            end
            Gamma = [gamma_min gamma];
            Jt = [J Jcross];
            [~,mu] = min(Jt);
            gold = (sqrt(5)+1)/2 ;
            while (gamma-gamma_min)>goldensearch_deltmax*(abs(deltmax))  && gamma > eps;
                stepmedr = gamma_min+(gamma-gamma_min)/gold;
                stepmedl = gamma_min+(stepmedr-gamma_min)/gold;
                dcross = obj.dm+stepmedr*D;
                K = obj.SD.computeKernelCombination(dcross,1:obj.SD.n); 
                SD = simpleData(K,obj.SD.trainlab);
                SD.kernelType='precomputed';
                SM = simpleModelENPSVM(SD,obj.C,obj.nbEig,SM.Il, SM.Iu, SM.Iw);
                SM.train;
                Jmedr = (-0.5*SM.alpha'*K([SM.Iw,SM.Iu],[SM.Iw,SM.Iu])*SM.alpha + sum(SM.alpha.*obj.SD.trainlab));
                dcross = obj.dm+stepmedl*D;
                K = obj.SD.computeKernelCombination(dcross,1:obj.SD.n);
                SD = simpleData(K,obj.SD.trainlab);
                SD.kernelType='precomputed';
                SM2 = simpleModelENPSVM(SD,obj.C,obj.nbEig,SM.Il, SM.Iu, SM.Iw);
                SM2.train;
                Jmedl = (-0.5*SM2.alpha'*K([SM2.Iw,SM2.Iu],[SM2.Iw,SM2.Iu])*SM2.alpha + sum(SM2.alpha.*obj.SD.trainlab));
                Gamma = [gamma_min stepmedl stepmedr gamma];
                Jt = [J Jmedl Jmedr Jcross];
                [~,mu] = min(Jt);
                switch mu
                    case 1
                        gamma = stepmedl;
                        Jcross = Jmedl;
                        
                        obj.Iu = SM2.Iu;
                        obj.Iw = SM2.Iw;
                        obj.Il = SM2.Il;
                        obj.alpha = SM2.alpha;
                        obj.b = SM2.b;
                        
                    case 2
                        gamma = stepmedr;
                        Jcross = Jmedr;
                        obj.Iu = SM.Iu;
                        obj.Iw = SM.Iw;
                        obj.Il = SM.Il;
                        obj.alpha = SM.alpha;
                        obj.b = SM.b;
                    case 3
                        gamma_min = stepmedl;
                        J = Jmedl;
                        obj.Iu = SM2.Iu;
                        obj.Iw = SM2.Iw;
                        obj.Il = SM2.Il;
                        obj.alpha = SM2.alpha;
                        obj.b = SM2.b;
                    case 4
                        gamma_min = stepmedr;
                        J = Jmedr;
                        obj.Iu = SM.Iu;
                        obj.Iw = SM.Iw;
                        obj.Il = SM.Il;
                        obj.alpha = SM.alpha;
                        obj.b = SM.b;
                end
                
                
                SM.clear;
                SM2.clear;
                
            end
            %---------------------------------
            % Final Updates
            %---------------------------------
            
            Jcross = Jt(mu) ;
            gamma = Gamma(mu) ;
            % Sigma update
            if Jcross < J ;
                obj.dm = obj.dm + gamma * D;
                
            end
            
        end
        
        
        function train(obj)
            goldensearch_deltmax = 0.1;
            iter = 0;
            oldDual = 10^12;
            while obj.converged==0 && iter<200
                dmOld = obj.dm;
                iter = iter+1;
                K = obj.SD.computeKernelCombination(obj.dm,1:obj.SD.n);
                SD = simpleData(K,obj.SD.trainlab);
                SD.kernelType='precomputed';
                SM = simpleModelENPSVM(SD,obj.C,obj.nbEig,obj.Il, obj.Iu, obj.Iw);
                SM.train;
                J = (-0.5*(SM.alpha)'*K([SM.Iw,SM.Iu],[SM.Iw,SM.Iu])*...
                    (SM.alpha) + sum(SM.alpha.*obj.SD.trainlab));
                
                obj.Iu = SM.Iu;
                obj.Iw = SM.Iw;
                obj.Il = SM.Il;
                obj.alpha = SM.alpha;
                obj.b = SM.b;
                [Jcross] = obj.updateGrads(SM,J, goldensearch_deltmax);
                
                obj.dm(obj.dm<10^-8) = 0;
                obj.dm = obj.dm/(sum(obj.dm));
                
                if max(abs(obj.dm-dmOld))<10^-8 && goldensearch_deltmax > 10^-8
                    goldensearch_deltmax=goldensearch_deltmax/10;
                elseif goldensearch_deltmax~=0.1
                    goldensearch_deltmax = goldensearch_deltmax*10;
                end;
                
                for m=1:length(obj.dm)
                    dJ(m) = -0.5*(obj.alpha)'*(obj.SD.kernelCachePart(m,[obj.Iw,obj.Iu]))*(obj.alpha);
                end
                
                dualGap = (max(-dJ)+Jcross-sum(obj.alpha.*obj.SD.trainlab));
                if (abs(dualGap)<obj.kktgap || abs(oldDual-dualGap)<obj.kktgap)
                    obj.converged=1;
                else
                    oldDual = dualGap;
                end
                
                fprintf('%d | %f | %f | %f \n',iter,dualGap,-Jcross+sum(obj.alpha.*obj.SD.trainlab),max(-dJ));
            end
        end
        
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelSVDD:print2D','must be 2D data');
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
            
            
            [ypred] = obj.test(xtest,ones(size(xtest,2),1));
            
            ypredmat=reshape(ypred,na,nb);
            
            plot(obj.SD.trainvec(1,obj.SD.trainlab==-1),obj.SD.trainvec(2,obj.SD.trainlab==-1),'k+');
            plot(obj.SD.trainvec(1,obj.SD.trainlab==1),obj.SD.trainvec(2,obj.SD.trainlab==1),'r+');
            hold on;
            plot(obj.SV(1,:),obj.SV(2,:),'om','MarkerSize',10,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ 0 0],'g');
            set(h,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[-1 1]);
            colorbar;
            set(h,'LineWidth',1);
            drawnow;
            
        end
        
        
        function [ypred perf] = test(obj,testdata,testlabel)
            
            
            nb = 0;
            for i=1:length(obj.SD.kernelType)
                if (strcmp(obj.SD.kernelType{i}{2},'precomputed'))
                    f = zeros(length([obj.Iw,obj.Iu]),length(testlabel));
                    for j = 1:length(obj.SD.kernelParam{i})
                        nb = nb + 1;
                        if obj.dm(nb)>0
                            f = f+ obj.dm(nb)*testdata{i}{j};
                        end
                    end
                    
                else
                    if (strcmp(obj.SD.kernelType{i}{2},'all'))
                        for j = 1:length(obj.SD.kernelParam{i})
                            nb = nb + 1;
                            type{nb} = obj.SD.kernelType{i}{1};
                            var{nb} = ':';
                            param{nb} = obj.SD.kernelParam{i}(j);
                        end
                    elseif (strcmp(obj.SD.kernelType{i}{2},'single'))
                        for k=1:size(obj.SD.trainvec,1)
                            for j=1:length(obj.SD.kernelParam{i})
                                nb = nb +1;
                                type{nb} = obj.SD.kernelType{i}{1};
                                var{nb} = k;
                                param{nb} = obj.SD.kernelParam{i}(j);
                            end
                        end
                    end
                    
                    f = zeros(length([obj.Iw,obj.Iu]),length(testlabel));
                    for pp=1:length(obj.dm)
                        if obj.dm(pp)>0
                            f = f+ obj.dm(pp)*kernelTest(type{pp},param{pp},obj.SD.trainvec(var{pp},[obj.Iw,obj.Iu]), ...
                                abs(obj.SD.trainlab([obj.Iw,obj.Iu])), testdata(var{pp},:), abs(testlabel))';
                        end
                    end
                end
                
            end
            ypred = (obj.alpha)'*f+obj.b;
            perf = 100*length(find(sign(ypred')==testlabel))/length(testlabel);
            
        end
        
    end
    
end
