classdef simpleMKLData < handle
    % SIMPLEMKLDATA This class contains all data that are used in simpleMKLModel
    % solvers. It stores initial data (or a pre-computed kernel), computes
    % the kernel matrix on-demand with a (simple) cache system.
    % See also : simpleMKLModel
    
    % Create by Galle Loosli
    % 29/11/2013
    % Last Modified 29/11/2013 by G.L.
    
    properties
        trainvec; % stored as column individuals
        trainlab; % stored as column
        kernelType = 'rbf';
        kernelParam = 1;
        kernelCache; % computed on the fly or pre-computed
        n;
        nbKernels = 1;
    end
    
    properties (SetAccess = private)
        indx; % cached rows
        relIndx; % position of cached rows
        premult;
    end
    
    
    methods
        % contructor
        function SD = simpleMKLData(tt,tl)
            assert(nargin==2,'simpleData:wrongInput','SimpleData needs training points or pre-computed kernel matrix and associated labels : two arguments');
            assert((size(tl,1)==size(tt,2) && size(tl,2)==1) || (iscell(tt) && size(tt{1},1)==size(tl,1)) ,'simpleData:wrongInput','Labels should be in a column vector, training point stored as columns');
            if iscell(tt)
                SD.trainlab = tl;
                SD.relIndx = zeros(1,size(tt{1},2));
                SD.n = size(tt{1},2);
                SD.kernelCache = tt;
                SD.nbKernels = size(tt,2);
                SD.kernelType='precomputed';
            else
                SD.trainvec = tt;
                SD.trainlab = tl;
                SD.relIndx = zeros(1,size(tt,2));
                SD.n = size(tt,2);
                if not(isstruct(tt))
                    SD.premult = tt.*tt;
                end
                for pp=1:length(SD.kernelParam)
                    SD.kernelCache{pp} = zeros(SD.n,min(500,SD.n));
                end
                SD.nbKernels = length(SD.kernelParam);
            end
        end
        
        function set.kernelType(obj,type)
            obj.kernelType = type;
        end
        
        function type = get.kernelType(obj)
            type = obj.kernelType;
        end
        
        function  set.kernelParam(obj,param)
            obj.kernelParam = param;
            %obj.nbKernels = length(param);
        end
        
        function param = get.kernelParam(obj)
            param = obj.kernelParam;
        end
        
        function K = computeKernelCombination(obj,dm,list)
            K = zeros(length(list));
            for pp=1:length(dm)
                K = K + dm(pp)*obj.kernelCachePart(pp,list);
            end
        end
        
        function K = kernelCachePart(obj,m,list)
            K = obj.kernelCache{m}(list,list);
        end
        
        function indd = getKernelValues(obj,indx,indy)
            %a = setdiff(indx, obj.indx);
            a = indx(~(ismember(indx,obj.indx)));
            if ~isempty(a)
                obj.computeKernel(a);
                obj.updateCache();
            end
            %K = obj.kernelCache(indy,obj.relIndx(indx));
            indd = obj.relIndx(indx);
        end
        
        function obj = computeKernel(obj,ind)
            
            if strcmp(obj.kernelType,'precomputed')
                
            else
                nb = 0;
                for i=1:length(obj.kernelParam)
                    if (strcmp(obj.kernelType{i}{2},'all'))
                        nb = nb + length(obj.kernelParam{i});
                    elseif (strcmp(obj.kernelType{i}{2},'single'))
                        nb = nb + length(obj.kernelParam{i})*size(obj.trainvec,1);
                    elseif (strcmp(obj.kernelType{i}{2},'precomputed'))
                        nb = nb+1;
                    end
                end
                obj.nbKernels = nb;
                
                relInd = length(obj.indx)+1:length(obj.indx)+length(ind);
                kernelLab(obj,ind,relInd);
                obj.relIndx(ind) = relInd;
                obj.indx = [obj.indx,ind];
            end
        end
        
        
        
        function indd = getDocKernelValues(obj,indx,indy,Q,lambda)
            %a = setdiff(indx, obj.indx);
            a = indx(~(ismember(indx,obj.indx)));
            if ~isempty(a)
                obj.computeDocKernel(a,Q,lambda);
                obj.updateCache();
            end
            %K = obj.kernelCache(indy,obj.relIndx(indx));
            indd = obj.relIndx(indx);
        end
        
        function obj = computeDocKernel(obj,ind,Q,lambda)
            relInd = length(obj.indx)+1:length(obj.indx)+length(ind);
            obj.kernelDocLab(ind,Q,lambda,relInd);
            obj.relIndx(ind) = relInd;
            obj.indx = [obj.indx,ind];
        end
        
        function obj = kernelLab(obj,ind,relInd)
            
            
            
            pp = 0;
            for nk = 1:length(obj.kernelType)
                if (strcmp(obj.kernelType{nk}(2),'all'))
                    if strcmp(obj.kernelType{nk}(1),'rbf')
                        dot_b = sum(obj.premult(:,ind),1);
                        dot_a = sum(obj.premult,1)';
                        unitvec = ones(obj.n,1);
                        for ppi=1:length(obj.kernelParam{nk})
                            pp = pp +1;
                            obj.kernelCache{pp}(:,relInd) =   obj.trainvec'*obj.trainvec(:,ind);
                            for i=1:length(ind)
                                a = obj.kernelParam{nk}(ppi) * ...
                                    (2 * obj.kernelCache{pp}(:,relInd(i)) - dot_a - dot_b(i) * unitvec);
                                obj.kernelCache{pp}(:,relInd(i)) = obj.trainlab(ind(i)) * (exp(a) .* obj.trainlab);
                            end
                        end
                    end
                    
                    if strcmp(obj.kernelType{nk}(1),'poly')
                        for ppi=1:length(obj.kernelParam{nk})
                            pp = pp +1;
                            obj.kernelCache{pp}(:,relInd) = (obj.trainvec' * obj.trainvec(:,ind) +...
                                obj.kernelParam{nk}(ppi,2)).^obj.kernelParam{nk}(ppi,1);
                            for i=1:length(ind)
                                obj.kernelCache{pp}(:,relInd(i)) = obj.trainlab(ind(i)) * (obj.kernelCache{pp}(:,relInd(i)) .* obj.trainlab);
                            end
                        end
                    end
                end
                if (strcmp(obj.kernelType{nk}(2),'single'))
                    for var = 1:size(obj.trainvec,1)
                        if strcmp(obj.kernelType{nk}{1},'rbf')
                            dot_b = sum(obj.premult(var,ind),1);
                            dot_a = sum(obj.premult(var,:),1)';
                            unitvec = ones(obj.n,1);
                            for ppi=1:length(obj.kernelParam{nk})
                                pp = pp+1;
                                obj.kernelCache{pp}(:,relInd) =   obj.trainvec(var,:)'*obj.trainvec(var,ind);
                                for i=1:length(ind)
                                    a = obj.kernelParam{nk}(ppi) * ...
                                        (2 * obj.kernelCache{pp}(:,relInd(i)) - dot_a - dot_b(i) * unitvec);
                                    obj.kernelCache{pp}(:,relInd(i)) = obj.trainlab(ind(i)) * (exp(a) .* obj.trainlab);
                                end
                            end
                        end
                        
                        if strcmp(obj.kernelType{nk}(1),'poly')
                            for ppi=1:length(obj.kernelParam{nk})
                                pp = pp +1;
                                obj.kernelCache{pp}(:,relInd) = (obj.trainvec(var,:)' * obj.trainvec(var,ind) +...
                                    obj.kernelParam{nk}(ppi,2)).^obj.kernelParam{nk}(ppi,1);
                                for i=1:length(ind)
                                    obj.kernelCache{pp}(:,relInd(i)) = obj.trainlab(ind(i)) * (obj.kernelCache{pp}(:,relInd(i)) .* obj.trainlab);
                                end
                            end
                        end
                    end
                end
                if (strcmp(obj.kernelType{nk}(2),'precomputed'))
                    for ppi=1:length(obj.kernelParam{nk})
                        pp = pp+1;
                        obj.kernelCache{pp} = obj.kernelParam{nk}{ppi};
                    end
                    
                end
            end
            
            if strcmp(obj.kernelType,'poly')
                assert(length(obj.kernelParam)==3);
                obj.kernelCache(:,relInd) = (obj.kernelParam(2) * obj.trainvec' * obj.trainvec(:,ind) +...
                    obj.kernelParam(3)).^obj.kernelParam(1);
                for i=1:length(ind)
                    obj.kernelCache(:,relInd(i)) = obj.trainlab(ind(i)) * (obj.kernelCache(:,relInd(i)) .* obj.trainlab);
                end
            end
            
            
            if strcmp(obj.kernelType,'tanh')
                assert(length(obj.kernelParam)==2);
                obj.kernelCache(:,relInd) = tanh(obj.kernelParam(1) * obj.trainvec' * obj.trainvec(:,ind) +...
                    obj.kernelParam(2));
                for i=1:length(ind)
                    obj.kernelCache(:,relInd(i)) = obj.trainlab(ind(i)) * (obj.kernelCache(:,relInd(i)) .* obj.trainlab);
                end
            end
            
            
            if strcmp(obj.kernelType,'precomputed')
                obj.kernelCache = obj.trainvec.*(obj.trainlab*obj.trainlab');
            end
            
            if strcmp(obj.kernelType,'graph')
                assert(length(obj.kernelParam)==3);
                for i=1:length(ind)
                    for j=1:length(obj.trainlab)
                        nbetiqnoeuds=size(obj.trainvec(relInd(i)).etiquetteNoeud,2);
                        nbetiqarcs=size(obj.trainvec(relInd(i)).edgeLabel(1,:),2);
                        sigma=ones(1,nbetiqnoeuds+nbetiqarcs); % general case
                        sigma(1,1:nbetiqnoeuds) = obj.kernelParam(1);
                        sigma(1,nbetiqnoeuds+1:end)=obj.kernelParam(2);
                        obj.kernelCache(i,relInd(j)) = obj.trainlab(ind(i))*obj.trainlab(relInd(j))*Kchemin(obj.trainvec(ind(i)),obj.trainvec(j),sigma,obj.kernelParam(3));
                    end
                end
            end
            
            
            if strcmp(obj.kernelType,'graphMax')
                assert(length(obj.kernelParam)==3);
                for i=1:length(ind)
                    for j=1:length(obj.trainlab)
                        nbetiqnoeuds=size(obj.trainvec(relInd(i)).etiquetteNoeud,2);
                        nbetiqarcs=size(obj.trainvec(relInd(i)).edgeLabel(1,:),2);
                        sigma=ones(1,nbetiqnoeuds+nbetiqarcs); % general case
                        sigma(1,1:nbetiqnoeuds) = obj.kernelParam(1);
                        sigma(1,nbetiqnoeuds+1:end)=obj.kernelParam(2);
                        obj.kernelCache(i,relInd(j)) = obj.trainlab(ind(i))*obj.trainlab(relInd(j))*KcheminMax(obj.trainvec(ind(i)),obj.trainvec(j),sigma,obj.kernelParam(3));
                    end
                end
            end
        end
        
        function obj = kernelDocLab(obj,ind,Q,lambda,relInd)
            if strcmp(obj.kernelType,'rbf')
                assert(length(obj.kernelParam)==1);
                dot_b = sum(obj.premult(:,ind),1);
                dot_a = sum(obj.premult,1)';
                unitvec = ones(obj.n,1);
                A = lambda/(1+Q*lambda);
                obj.kernelCache(:,relInd) =   obj.trainvec'*obj.trainvec(:,ind);
                if lambda==0
                    for i=1:length(ind)
                        obj.kernelCache(:,relInd(i)) =(obj.trainlab(ind(i))==obj.trainlab) .* (exp(obj.kernelParam(1) * ...
                            (2 * obj.kernelCache(:,relInd(i)) - dot_a - dot_b(i) * unitvec)));
                    end
                else
                    B = 1/lambda + (Q-1);
                    for i=1:length(ind)
                        obj.kernelCache(:,relInd(i)) =(1-(B-1)*(obj.trainlab(ind(i))==obj.trainlab)) .* (-A*exp(obj.kernelParam(1) * ...
                            (2 * obj.kernelCache(:,relInd(i)) - dot_a - dot_b(i) * unitvec)));
                    end
                end
            end
            
            if strcmp(obj.kernelType,'poly')
                assert(length(obj.kernelParam)==3);
                obj.kernelCache(:,relInd) = (obj.kernelParam(2) * obj.trainvec' * obj.trainvec(:,ind) +...
                    obj.kernelParam(3)).^obj.kernelParam(1);
                A= lambda/(1+Q*lambda);
                if lambda==0
                    for i=1:length(ind)
                        obj.kernelCache(:,relInd(i)) = obj.trainlab(ind(i)) * (obj.kernelCache(:,relInd(i)) .* obj.trainlab);
                    end
                else
                    B= 1/lambda + (Q-1);
                    for i=1:length(ind)
                        obj.kernelCache(:,relInd(i)) = -B.^(obj.trainlab(ind(i))==obj.trainlab).*(-A*obj.kernelCache(:,relInd(i)));
                    end
                end
                
            end
            
            if strcmp(obj.kernelType,'tanh')
                assert(length(obj.kernelParam)==2);
                obj.kernelCache(:,relInd) = tanh(obj.kernelParam(1) * obj.trainvec' * obj.trainvec(:,ind) +...
                    obj.kernelParam(2));
                A= lambda/(1+Q*lambda);
                if lambda==0
                    for i=1:length(ind)
                        obj.kernelCache(:,relInd(i)) = obj.trainlab(ind(i)) * (obj.kernelCache(:,relInd(i)) .* obj.trainlab);
                    end
                else
                    B= 1/lambda + (Q-1);
                    for i=1:length(ind)
                        obj.kernelCache(:,relInd(i)) = -B.^(obj.trainlab(ind(i))==obj.trainlab).*(-A*obj.kernelCache(:,relInd(i)));
                    end
                end
                
            end
            
            if strcmp(obj.kernelType,'precomputed')
                A= lambda/(1+Q*lambda);
                if lambda==0
                    for i=1:length(ind)
                        obj.kernelCache(:,relInd(i)) = obj.trainlab(ind(i)) * (obj.kernelCache(:,relInd(i)) .* obj.trainlab);
                    end
                else
                    B= 1/lambda + (Q-1);
                    for i=1:length(ind)
                        obj.kernelCache(:,relInd(i)) = -B.^(obj.trainlab(ind(i))==obj.trainlab).*(-A*obj.kernelCache(:,relInd(i)));
                    end
                end
            end
            
            
            
            
            if strcmp(obj.kernelType,'graph')
                assert(length(obj.kernelParam)==3);
                A= lambda/(1+Q*lambda);
                if lambda==0
                    for i=1:length(ind)
                        for j=1:length(obj.trainlab)
                            nbetiqnoeuds=size(obj.trainvec(i).etiquetteNoeud,2);
                            nbetiqarcs=size(obj.trainvec(i).edgeLabel(1,:),2);
                            sigma=ones(1,nbetiqnoeuds+nbetiqarcs); % general case
                            sigma(1,1:nbetiqnoeuds) = obj.kernelParam(1);
                            sigma(1,nbetiqnoeuds+1:end)=obj.kernelParam(2);
                            obj.kernelCache(j,relInd(i)) =(obj.trainlab(ind(i))==obj.trainlab(j)) .* ...
                                (-A*Kchemin(obj.trainvec(ind(i)),obj.trainvec(j),sigma,obj.kernelParam(3)));
                        end
                    end
                else
                    B = 1/lambda + (Q-1);
                    for i=1:length(ind)
                        for j=1:length(obj.trainlab)
                            nbetiqnoeuds=size(obj.trainvec(relInd(i)).etiquetteNoeud,2);
                            nbetiqarcs=size(obj.trainvec(relInd(i)).edgeLabel(1,:),2);
                            sigma=ones(1,nbetiqnoeuds+nbetiqarcs); % general case
                            sigma(1,1:nbetiqnoeuds) = obj.kernelParam(1);
                            sigma(1,nbetiqnoeuds+1:end)=obj.kernelParam(2);
                            obj.kernelCache(j,relInd(i)) =(1-(B-1)*(obj.trainlab(ind(i))==obj.trainlab(j))) .* ...
                                (-A*Kchemin(obj.trainvec(ind(i)),obj.trainvec(j),sigma,obj.kernelParam(3)));
                        end
                    end
                end
                
            end
            if strcmp(obj.kernelType,'graphMax')
                assert(length(obj.kernelParam)==3);
                A= lambda/(1+Q*lambda);
                if lambda==0
                    for i=1:length(ind)
                        for j=1:length(obj.trainlab)
                            nbetiqnoeuds=size(obj.trainvec(i).etiquetteNoeud,2);
                            nbetiqarcs=size(obj.trainvec(i).edgeLabel(1,:),2);
                            sigma=ones(1,nbetiqnoeuds+nbetiqarcs); % general case
                            sigma(1,1:nbetiqnoeuds) = obj.kernelParam(1);
                            sigma(1,nbetiqnoeuds+1:end)=obj.kernelParam(2);
                            obj.kernelCache(j,relInd(i)) =(obj.trainlab(ind(i))==obj.trainlab(j)) .* ...
                                (-A*KcheminMax(obj.trainvec(ind(i)),obj.trainvec(j),sigma,obj.kernelParam(3)));
                        end
                    end
                else
                    B = 1/lambda + (Q-1);
                    for i=1:length(ind)
                        for j=1:length(obj.trainlab)
                            nbetiqnoeuds=size(obj.trainvec(relInd(i)).etiquetteNoeud,2);
                            nbetiqarcs=size(obj.trainvec(relInd(i)).edgeLabel(1,:),2);
                            sigma=ones(1,nbetiqnoeuds+nbetiqarcs); % general case
                            sigma(1,1:nbetiqnoeuds) = obj.kernelParam(1);
                            sigma(1,nbetiqnoeuds+1:end)=obj.kernelParam(2);
                            obj.kernelCache(j,relInd(i)) =(1-(B-1)*(obj.trainlab(ind(i))==obj.trainlab(j))) .* ...
                                (-A*KcheminMax(obj.trainvec(ind(i)),obj.trainvec(j),sigma,obj.kernelParam(3)));
                        end
                    end
                end
                
            end
        end
        
        
        
        function updateCache(obj)
            if length(obj.indx)+10>size(obj.kernelCache,2)
                obj.kernelCache = [obj.kernelCache,zeros(obj.n,100)];
            end
        end
        
        
        function clear(obj)
            obj.indx = [];
            obj.relIndx = obj.relIndx*0;
            if strcmp(obj.kernelType,'precomputed')
                obj.indx = 1:size(obj.trainvec,1);
                obj.relIndx = 1:size(obj.trainvec,1);
                obj.kernelCache = obj.trainvec.*(obj.trainlab*obj.trainlab');
            else
                obj.kernelCache = zeros(obj.n,500);
            end
        end
        
    end
    
end

