classdef simpleMKLModelSVR < simpleMKLModel
    % SIMPLEMODELCSVM This class inherits from simpleModel. It contains
    % particular methods to solve binary-C-SVM.
    % This class support C-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    %  9/01/2014
    % Last Modified  9/01/2014 by G.L.
    properties
        etube; % epsilon tube
        targets;
    end
    
    
    methods
        % contructor
        function SM = simpleMKLModelSVR(SD,C,etube,Il,Iu,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelCSVM object
            
            assert(nargin==3 || nargin==6,'simpleModelSVR:wrongInput','Requires 3 or 6 input arguments');
            if nargin<2
                error('not enough input');
            else
                
                %SM.SD = simpleMKLData([SD.trainvec, SD.trainvec],[ones(size(SD.trainlab));-ones(size(SD.trainlab))]);
                %SM.SD.kernelType = SD.kernelType;
                %SM.SD.kernelParam = SD.kernelParam; 
                SM.SD = SD;
                l = SD.trainlab;
                SM.SD.trainlab = ones(size(SD.trainlab));
                SM.SD.getKernelValues(1:SM.SD.n);
                SM.targets = [SD.trainlab-etube; -SD.trainlab-etube];
                SM.SD.trainlab = l; 
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                if nargin==6
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [1,2];
                    SM.Iu = [];
                    SM.Il = 1:2*SM.SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(length(SM.Iw),1);
                end
                SM.dm = ones(1,SD.nbKernels)./SD.nbKernels;
                SM.Slacks = C(1)*ones(size(SM.SD.trainlab));
                SM.activated = 0;
                SM.converged = 0;
                SM.etube = etube;
            end

        end
        
        
        function val = get.etube(obj)
            val = obj.etube;
        end
        
        
        function set.etube(obj,val)
            obj.etube = val;
        end
        
        
        function val = get.targets(obj)
            val = obj.targets;
        end
        
        
        function set.targets(obj,val)
            obj.targets = val;
        end
        
        function [Jcross] = updateGrads(obj,SM,J,goldensearch_deltmax)
            [~,mu] = max(obj.dm);
            
            yl = [obj.SD.trainlab;obj.SD.trainlab];
            
            for m=1:length(obj.dm)
                Kt = [obj.SD.kernelCachePart(m,:),obj.SD.kernelCachePart(m,:);obj.SD.kernelCachePart(m,:),obj.SD.kernelCachePart(m,:)];
                dJ(m) = -0.5*(obj.alpha)'*Kt([obj.Iw,obj.Iu],[obj.Iw,obj.Iu])*(obj.alpha);
            end
            NormGrad = dJ*dJ';
            dJ=dJ/sqrt(NormGrad);
            
            dJ = dJ - dJ(mu);
            D = -dJ.*( (obj.dm>0) | (dJ<0));
            D(mu) = -sum(D);
            
            gamma_min = 0;
            Jcross = 0;
            
            ind = find(D<0);
            gamma = min(-(obj.dm(ind))./D(ind));
            deltmax = gamma;
            if isempty(gamma) || gamma==0
                obj.dm = obj.dm ;
                return
            end
            if gamma > 0.1
                gamma=0.1;
            end;
            
            while (J>Jcross)
                dcross = obj.dm+gamma*D;
                K = obj.SD.computeKernelCombination(dcross,1:obj.SD.n);
                SD = simpleData(K,ones(size(obj.SD.trainlab)));
                SD.trainlab = obj.SD.trainlab;
                SD.kernelType = 'precomputed';
                SM.clear;
                SM = simpleModelSVR(SD,obj.C,obj.etube,SM.Il, SM.Iu, SM.Iw);
                SM.train;
                Jcross = (-0.5*abs(SM.alpha)'*SM.SD.kernelCache([SM.Iw,SM.Iu],[SM.Iw,SM.Iu])*abs(SM.alpha) + obj.etube*sum(abs(SM.alpha)) + SM.alpha'*yl([SM.Iw,SM.Iu]) )
                if (Jcross<J)
                    J = Jcross;
                    obj.dm = dcross;
                    D = D.*(dcross>10^-8 | D>0);
                    D(mu) =  - sum(D([1:mu-1,mu+1:end]));
                    ind = find(D<0);
                    obj.Iu = SM.Iu;
                    obj.Iw = SM.Iw;
                    obj.Il = SM.Il;
                    obj.alpha = SM.alpha;
                    obj.b = SM.b;
                    
                    if ~isempty(ind)
                        gamma = min(-(obj.dm(ind))./D(ind));
                        deltmax = gamma;
                        Jcross = 0;
                    else
                        gamma = 0;
                        deltmax = 0;
                    end
                    if gamma > 1
                        gamma=1;
                    end;
                end
                
            end
            Gamma = [gamma_min gamma];
            Jt = [J Jcross];
            [~,mu] = min(Jt);
            gold = (sqrt(5)+1)/2 ;
            while (gamma-gamma_min)>goldensearch_deltmax*(abs(deltmax))  && gamma > eps;
                stepmedr = gamma_min+(gamma-gamma_min)/gold;
                stepmedl = gamma_min+(stepmedr-gamma_min)/gold;
                dcross = obj.dm+stepmedr*D;
                K = obj.SD.computeKernelCombination(dcross,1:obj.SD.n);
                SD = simpleData(K,ones(size(obj.SD.trainlab)));
                SD.trainlab = obj.SD.trainlab;
                SD.kernelType = 'precomputed';
                SM = simpleModelSVR(SD,obj.C,obj.etube,SM.Il, SM.Iu, SM.Iw);
                SM.train;
                Jmedr =( -0.5*abs(SM.alpha)'*SM.SD.kernelCache([SM.Iw,SM.Iu],[SM.Iw,SM.Iu])*abs(SM.alpha) + obj.etube*sum(abs(SM.alpha)) + SM.alpha'*yl([SM.Iw,SM.Iu])) ;
                dcross = obj.dm+stepmedl*D;
                K = (obj.SD.trainlab*obj.SD.trainlab').*obj.SD.computeKernelCombination(dcross,1:obj.SD.n);
                SD = simpleData(K,ones(size(obj.SD.trainlab)));
                SD.trainlab = obj.SD.trainlab;
                SD.kernelType = 'precomputed';
                SM2 = simpleModelSVR(SD,obj.C,obj.etube,SM.Il, SM.Iu, SM.Iw);
                SM2.train;
                Jmedl = (-0.5*(abs(SM2.alpha))'*SM2.SD.kernelCache([SM2.Iw,SM2.Iu],[SM2.Iw,SM2.Iu])*abs(SM2.alpha) + obj.etube*sum(abs(SM2.alpha)) + SM2.alpha'*yl([SM2.Iw,SM2.Iu]) );
                Gamma = [gamma_min stepmedl stepmedr gamma];
                Jt = [J Jmedl Jmedr Jcross];
                [~,mu] = min(Jt);
                switch mu
                    case 1
                        gamma = stepmedl;
                        Jcross = Jmedl;
                        
                        obj.Iu = SM2.Iu;
                        obj.Iw = SM2.Iw;
                        obj.Il = SM2.Il;
                        obj.alpha = SM2.alpha;
                        obj.b = SM2.b;
                        
                    case 2
                        gamma = stepmedr;
                        Jcross = Jmedr;
                        obj.Iu = SM.Iu;
                        obj.Iw = SM.Iw;
                        obj.Il = SM.Il;
                        obj.alpha = SM.alpha;
                        obj.b = SM.b;
                    case 3
                        gamma_min = stepmedl;
                        J = Jmedl;
                        obj.Iu = SM2.Iu;
                        obj.Iw = SM2.Iw;
                        obj.Il = SM2.Il;
                        obj.alpha = SM2.alpha;
                        obj.b = SM2.b;
                    case 4
                        gamma_min = stepmedr;
                        J = Jmedr;
                        obj.Iu = SM.Iu;
                        obj.Iw = SM.Iw;
                        obj.Il = SM.Il;
                        obj.alpha = SM.alpha;
                        obj.b = SM.b;
                end
                
                
               
                SM.clear;
                SM2.clear; 
            end
            %---------------------------------
            % Final Updates
            %---------------------------------
            
            Jcross = Jt(mu) ;
            gamma = Gamma(mu);
            % Sigma update
            if Jcross < J ;
                obj.dm = obj.dm + gamma * D;
                
            end
            
        end
        
        
        
        
        function train(obj)
            goldensearch_deltmax = 0.1;
            iter = 0;
            yl = [obj.SD.trainlab;obj.SD.trainlab];
            while obj.converged==0 && iter<500
                dmOld = obj.dm;
                iter = iter+1;
                K = obj.SD.computeKernelCombination(obj.dm,1:obj.SD.n); 
                %min(min(K))
                SD = simpleData(K,ones(size(obj.SD.trainlab)));
                SD.trainlab = obj.SD.trainlab; 
                SD.kernelType = 'precomputed';
                SM = simpleModelSVR(SD,obj.C,obj.etube,obj.Il, obj.Iu, obj.Iw);
                SM.train;
                %J = -0.5*(abs(SM.alpha))'*SM.SD.kernelCache([SM.Iw,SM.Iu],[SM.Iw,SM.Iu])*(abs(SM.alpha)) + obj.etube*sum(abs(SM.alpha)) + SM.alpha'*yl([SM.Iw,SM.Iu]) 
                %[~, ii] = SM.test([K;K],obj.SD.trainlab)
                J = (-0.5*(abs(SM.alpha))'*SM.SD.kernelCache([SM.Iw,SM.Iu],[SM.Iw,SM.Iu])*(abs(SM.alpha)) + SM.alpha'*obj.targets([SM.Iw,SM.Iu]))
                
                obj.Iu = SM.Iu;
                obj.Iw = SM.Iw;
                obj.Il = SM.Il;
                obj.alpha = SM.alpha;
                obj.b = SM.b;
                [Jcross] = obj.updateGrads(SM,J, goldensearch_deltmax);
                
                obj.dm(obj.dm<10^-8) = 0;
                obj.dm = obj.dm/(sum(obj.dm));
                
                if max(abs(obj.dm-dmOld))<10^-8 && goldensearch_deltmax > 10^-8
                    goldensearch_deltmax=goldensearch_deltmax/10;
                elseif goldensearch_deltmax~=0.1
                    goldensearch_deltmax = goldensearch_deltmax*10;
                end;
                
                for m=1:length(obj.dm)
                    Kt = [obj.SD.kernelCachePart(m,:),obj.SD.kernelCachePart(m,:);obj.SD.kernelCachePart(m,:),obj.SD.kernelCachePart(m,:)];
                    dJ(m) = -0.5*(obj.alpha)'*Kt([obj.Iw,obj.Iu],[obj.Iw,obj.Iu])*(obj.alpha);
                end
                
                dualGap = (max(-dJ)+Jcross - abs(obj.alpha)'*obj.targets([obj.Iw,obj.Iu]))/Jcross
                if (abs(dualGap)<obj.kktgap)
                    obj.converged=1;
                end
                
                %fprintf('%d \t| %f | %d | %d \n',iter,dualGap,length(obj.Iw), length(obj.Iu));
            end
        end
        
        
        
        
      
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==1,'simpleModelCSVM:printD','must be 1D data');
            xtest=linspace(min(obj.SD.trainvec),max(obj.SD.trainvec),1000);
            obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
            
            [ypred] = obj.test(xtest,ones(size(xtest,2),1));
            
            hold off;
            plot(obj.SD.trainvec(1:obj.SD.n),obj.targets(1:obj.SD.n)+obj.etube,'-k');
            hold on;
            plot(obj.SD.trainvec(1:obj.SD.n),obj.targets(1:obj.SD.n),'-.k');
            plot(obj.SD.trainvec(1:obj.SD.n),-obj.targets(obj.SD.n+1:end),'-.k');
            tl = [obj.SD.trainlab;obj.SD.trainlab];
            plot(obj.SV,obj.targets([obj.Iw,obj.Iu]).*tl([obj.Iw,obj.Iu]),'om','MarkerSize',10,'LineWidth',2);
            plot(xtest,ypred,'-rx');
            
            drawnow;
            
        end
        
    
           
        function [ypred perf] = test(obj,testdata,testlabel)
            
            nb = 0;
            for i=1:length(obj.SD.kernelParam)
                if (strcmp(obj.SD.kernelType{i}{2},'all'))
                    for j = 1:length(obj.SD.kernelParam{i})
                        nb = nb + 1;
                        type{nb} = obj.SD.kernelType{i}{1};
                        var{nb} = ':';
                        param{nb} = obj.SD.kernelParam{i}(j);
                    end
                elseif (strcmp(obj.SD.kernelType{i}{2},'single'))
                    for k=1:size(obj.SD.trainvec,1)
                        for j=1:length(obj.SD.kernelParam{i})
                            nb = nb +1;
                            type{nb} = obj.SD.kernelType{i}{1};
                            var{nb} = k;
                            param{nb} = obj.SD.kernelParam{i}(j);
                        end
                    end
                end
            end
            
            f = zeros(length([obj.Iw,obj.Iu]),length(testlabel));
            
            for pp=1:length(obj.dm)
                tv = [obj.SD.trainvec,obj.SD.trainvec];
                if obj.dm(pp)>0
                    f = f+ obj.dm(pp)*kernelTest(type{pp},param{pp},tv(var{pp},[obj.Iw,obj.Iu]), ...
                        ones([obj.Iw,obj.Iu],1), testdata(var{pp},:), ones(size(testlabel)))';
                end
            end
            ypred = (obj.alpha)'*f+obj.b;
            perf = corr(testlabel,ypred');
            
            
        end
        
       
        
        
       
        
    end
    
end
