classdef simpleModelCSVMN < simpleModel
    % SIMPLEMODELCSVM This class inherits from simpleModel. It contains
    % particular methods to solve binary-C-SVM.
    % This class support C-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 16/02/2011
    % Last Modified 16/02/2011 by G.L.
    
    properties
       e; 
       olde;
    end
    
    methods
        function set.e(obj,val)
            obj.e = val;
        end
        
        function i = get.e(obj)
            i = obj.e;
        end
        function set.olde(obj,val)
            obj.olde = val;
        end
        
        function i = get.olde(obj)
            i = obj.olde;
        end
        % contructor
        function SM = simpleModelCSVMN(SD,C,Il,Iu,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelCSVM object
            
            assert(nargin==2 || nargin==5,'simpleModelCSVM:wrongInput','Requires 2 or 5 input arguments');
            if nargin<2
                error('not enough input');
            else
                SM.SD = SD;
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                if nargin==5
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [find(SD.trainlab==-1,1),find(SD.trainlab==1,1)];
                    SM.Iu = [];
                    SM.Il = 1:SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(2,1);
                end
                SM.labs = sort(unique(SD.trainlab));
                if length(C)==length(SD.trainlab)
                    SM.Slacks = C;
                elseif length(C)==length(SM.labs)
                    for i=1:length(SM.labs)
                        SM.Slacks(SD.trainlab==SM.labs(i)) = C(i);
                    end
                else
                    SM.Slacks = C(1)*ones(size(SD.trainlab));
                    
                end
                SM.activated = 0;
                SM.converged = 0;
                SM.e = ones(size(SD.trainlab));
                
            end
        end
        

        
        function solveLinearSystem(obj)
            ce = ones(1,length(obj.Iw));
            be = 0;
            indd = obj.SD.getKernelValues(obj.Iw,obj.Iu);
            if ~isempty(obj.Iu)
                ce = ce - (obj.Slacks(obj.Iu))'*obj.SD.kernelCache(obj.Iu,indd);
                be = be - (abs(obj.Slacks(obj.Iu))'*obj.SD.trainlab(obj.Iu));
            end
            if obj.optimType==1
                [L,U] = lu(obj.SD.kernelCache(obj.Iw,indd));
                M = (obj.e(obj.Iw).*obj.SD.trainlab(obj.Iw))'*(U\(L\obj.SD.trainlab(obj.Iw)));
                obj.b = M\((obj.e(obj.Iw).*obj.SD.trainlab(obj.Iw))'*(U\(L\ce'))-be);
                obj.alpha = U\(L\(ce'-obj.SD.trainlab(obj.Iw)*obj.b));
                %obj.e(obj.Iw(obj.oldalpha==0)) = sign(obj.alpha(obj.oldalpha==0))';
            else
                error('not implemented yet');
            end
            %sum((obj.e(obj.Iw).*obj.alpha).*obj.SD.trainlab(obj.Iw))...
            %    +sum(obj.Slacks(obj.Iu).*obj.SD.trainlab(obj.Iu))
        end
        
          
        function train(obj)
            iter = 0;
            while obj.converged==0 && iter<1000*length(obj.SD.n)
                iter = iter+1;
                
                obj.solveLinearSystem();
                obj.print2D(); 
                %keyboard
                obj.activateBoxConstraints();
                if (obj.activated==0)
                    obj.relaxBoxConstraints();
                end
            end
            if obj.converged==0
                disp('No convergence');
                obj.solveLinearSystem();
                obj.converged = 1;
            end
            obj.finish();
            %obj.print2D();
            
        end
        
                
        function activateBoxConstraints(obj)
            v = mean(abs(obj.alpha));
            seuil = 0.1*v
            if (~isempty(find(abs(obj.alpha)<seuil, 1)) || ~isempty(find(abs(obj.alpha)>obj.Slacks(obj.Iw), 1))) && length(obj.Iw)>2
                dir = obj.alpha - obj.oldalpha;
                indad = find(abs(obj.alpha) < seuil);
                indsup = find(abs(obj.alpha) > obj.Slacks(obj.Iw));
                [tI indmin] = min(abs(seuil*sign(obj.oldalpha(indad))-obj.oldalpha(indad)./dir(indad)));
                [tS indS] = min(abs(obj.Slacks(obj.Iw(indsup))-obj.oldalpha(indsup)./dir(indsup)));
                if isempty(tI) , tI = tS + 1; end;
                if isempty(tS) , tS = tI + 1; end;
                t = min(tI,tS);
                    %obj.oldalpha = obj.oldalpha +t*dir;
                if t==tI
                    
                    obj.oldalpha = obj.oldalpha +sign(obj.oldalpha(indad(indmin)))*t*dir;
                    %obj.oldalpha = obj.oldalpha - (-obj.oldalpha(indad(indmin))/dir(indad(indmin)))*dir;
                    obj.Il = sort([obj.Il , obj.Iw(indad(indmin))]);
                    %obj.e(obj.Iw(indad(indmin))) = -obj.e(obj.Iw(indad(indmin)));
                    obj.Iw(indad(indmin)) = [];
                    obj.oldalpha(indad(indmin)) = [];
                    obj.alpha(indad(indmin)) = [];
                else
                    obj.oldalpha = obj.oldalpha + sign(obj.oldalpha(indsup(indS)))*t*dir;
                    obj.Iu = sort([obj.Iu , obj.Iw(indsup(indS))]);
                    obj.Slacks(obj.Iw(indsup(indS))) = sign(obj.alpha(indsup(indS)))*obj.Slacks(obj.Iw(indsup(indS)));
                    obj.Iw(indsup(indS))= [];
                    obj.oldalpha(indsup(indS))= [];
                    obj.alpha(indsup(indS)) = [];
                end
                obj.activated = 1;
            else
                obj.activated = 0;
            end
            
        end
        
          
        function y = evaluateSolution(obj,candidates)
            indd = obj.SD.getKernelValues([obj.Iw,obj.Iu],candidates);
            [val,inds]= sort(indd);
            temp = [obj.alpha;obj.e(obj.Iu).*obj.Slacks(obj.Iu)];
            y = obj.SD.kernelCache(candidates,val)*temp(inds)+obj.SD.trainlab(candidates)*obj.b;
        end
        
        function reinit(obj)
            obj.Iw = [find(obj.SD.trainlab==-1,1),find(obj.SD.trainlab==1,1)];
            obj.Iu = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            
            if length(obj.Slacks)~=length(obj.SD.trainlab)
                if length(obj.C)==length(obj.SD.trainlab)
                    obj.Slacks = obj.C;
                elseif length(obj.C)==length(obj.labs)
                    for i=1:length(obj.labs)
                        obj.Slacks(obj.SD.trainlab==obj.labs(i)) = obj.C(i);
                    end
                else
                    obj.Slacks = obj.C(1)*ones(size(obj.SD.trainlab));
                    
                end
            end
            obj.activated = 0;
            obj.converged = 0;
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
            f = kernelEval(obj,xtest);
            if obj.converged==0
                ypred = ([obj.alpha;obj.Slacks(obj.Iu)].*obj.SD.trainlab([obj.Iw,obj.Iu]))'*f+obj.b;
            else
                ypred = obj.alpha'*f+obj.b;
            end
            ypredmat=reshape(ypred,na,nb);
            
            hold off;
            plot(obj.SD.trainvec(1,:),obj.SD.trainvec(2,:),'k+');
            hold on;
            plot(obj.SV(1,:),obj.SV(2,:),'om','MarkerSize',10,'LineWidth',2);        
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ 0 0],'g');
            set(h,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ -1 1],'y');
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        function relaxBoxConstraints(obj)
            xt = zeros(obj.SD.n,1);
            xt(obj.Iw) = obj.alpha;
            xt(obj.Iu) = obj.e(obj.Iu).*obj.Slacks(obj.Iu);
            S = 500;
            Found=0;
            nbLoop = ceil(length(obj.Il)/S);
            z = 0;
            if not(isempty(obj.Il))
                while (z<nbLoop && Found==0)
                    z = z+1;
                    nul = obj.evaluateSolution(obj.Il((obj.iloop-1)*S+1:min(obj.iloop*S,end)))-1;
                    [mm mpos]=min(nul);
                    if (mm<-obj.kktgap)
                        Found=1;
                        mpos = mpos+(obj.iloop-1)*S;
                    else
                        mm = inf; mpos = [];
                    end
                    %obj.iloop= obj.iloop+1;
                    obj.iloop = mod(obj.iloop,nbLoop)+1;
                    %keyboard
                end
            else
                mm = inf; mpos = [];
            end
            if not(isempty(obj.Iu))
                nuu = obj.evaluateSolution(obj.Iu)-1;
                [mmS mposS]=min(-nuu);
            else
                mmS = inf; mposS = [];
            end
            
            if (~isempty(mmS) && mmS < -obj.kktgap) || (~isempty(mm) && mm < -obj.kktgap)
                if isempty(mmS) || mm < mmS
                    obj.Iw = sort([obj.Iw , obj.Il(mpos)]);
                    obj.oldalpha = xt(obj.Iw);
                   
                    obj.Il(mpos) = [];
                    
                    
                else
                    obj.Iw = sort([obj.Iw , obj.Iu(mposS)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.Iu(mposS)=[];
                end
            else
                obj.converged = 1;
            end
        end
        
        function finish(obj)
            if ~isempty(obj.SD.trainvec)
                obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
                
            else
                %disp('no training data can be stored for later use :
                %pre-computed matrix');
            end
            obj.alpha = [obj.alpha;obj.e(obj.Iu).*obj.Slacks(obj.Iu)].*obj.SD.trainlab([obj.Iw,obj.Iu]);
        end

        
        function print2DMulti(obj,SMt)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            assert(nargin==2);
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            if length(SMt)==length(unique(obj.SD.trainlab));
                ypred = obj.test1vsAll(SMt,xtest);
            else
                ypred = obj.test1vs1(SMt,xtest);
            end
            ypredmat=reshape(ypred,na,nb);
            %labs = unique(obj.SD.trainlab);
            hold on;
            [c,h] = contour(xtesta1,xtesta2,ypredmat);
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
    end
    
end
