classdef simpleModelDNPSVM < simpleModel
    % SIMPLEMODELCSVM This class inherits from simpleModel. It contains
    % particular methods to solve binary-C-SVM.
    % This class support C-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 16/02/2011
    % Last Modified 16/02/2011 by G.L.
    
    properties
        Iup;
        Iun;
    end
    
    
    methods
        % contructor
        function SM = simpleModelDNPSVM(SD,C,Il,Iup,Iun,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelCSVM object
            
            assert(nargin==2 || nargin==6,'simpleModelCSVM:wrongInput','Requires 2 or 6 input arguments');
            if nargin<2
                error('not enough input');
            else
                SM.SD = SD;
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                if nargin==6
                    SM.Il = Il;
                    SM.Iup = Iup;
                    SM.Iun = Iun;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [find(SD.trainlab==-1,1),find(SD.trainlab==1,1)];
                    SM.Iup = [];
                    SM.Iun = [];
                    SM.Il = 1:SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(2,1);
                end
                SM.labs = sort(unique(SD.trainlab));
                if length(C)==length(SD.trainlab)
                    SM.Slacks = C;
                elseif length(C)==length(SM.labs)
                    for i=1:length(SM.labs)
                        SM.Slacks(SD.trainlab==SM.labs(i)) = C(i);
                    end
                else
                    SM.Slacks = C(1)*ones(size(SD.trainlab));
                    
                end
                SM.activated = 0;
                SM.converged = 0;
            end
        end
        
        
        
        function set.Iup(obj,val)
            obj.Iup = val;
        end
        
        function val = get.Iup(obj)
            val = obj.Iup;
        end
        
        
        function set.Iun(obj,val)
            obj.Iun = val;
        end
        
        function val = get.Iun(obj)
            val = obj.Iun;
        end
        
        
        
        function solveLinearSystem(obj)
            ce = ones(1,length(obj.Iw));
            be = 0;
            indd = obj.SD.getKernelValues(obj.Iw,[obj.Iup,obj.Iun]);
            if ~isempty(obj.Iu)
                ce = ce - (obj.Slacks(obj.Iup))'*obj.SD.kernelCache(obj.Iup,indd)...
                    + (obj.Slacks(obj.Iun))'*obj.SD.kernelCache(obj.Iun,indd);
                be = be - obj.Slacks(obj.Iup)'*obj.SD.trainlab(obj.Iup) +...
                    obj.Slacks(obj.Iun)'*obj.SD.trainlab(obj.Iun);
            end
            if obj.optimType==1
                [Q,R] = qr(obj.SD.kernelCache(obj.Iw,indd)+obj.kktgap*eye(length(obj.Iw)));
                M = obj.SD.trainlab(obj.Iw)'*(Q*(R'\obj.SD.trainlab(obj.Iw)));
                obj.b = M\(obj.SD.trainlab(obj.Iw)'*(Q*(R'\ce'))-be);
                obj.alpha = Q*(R'\(ce'-obj.SD.trainlab(obj.Iw)*obj.b));
            else
                error('not implemented yet');
            end
        end
        
        
        function reinit(obj)
            obj.Iw = [find(obj.SD.trainlab==-1,1),find(obj.SD.trainlab==1,1)];
            obj.Iup = [];
            obj.Iun = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            
            if length(obj.Slacks)~=length(obj.SD.trainlab)
                if length(obj.C)==length(obj.SD.trainlab)
                    obj.Slacks = obj.C;
                elseif length(obj.C)==length(obj.labs)
                    for i=1:length(obj.labs)
                        obj.Slacks(obj.SD.trainlab==obj.labs(i)) = obj.C(i);
                    end
                else
                    obj.Slacks = obj.C(1)*ones(size(obj.SD.trainlab));
                    
                end
            end
            obj.activated = 0;
            obj.converged = 0;
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iup,obj.Iun]);
            f = kernelEval(obj,xtest);
            if obj.converged==0
                ypred = ([obj.alpha;obj.Slacks(obj.Iup);-obj.Slacks(obj.Iun)]...
                    .*obj.SD.trainlab([obj.Iw,obj.Iup,obj.Iun]))'*f+obj.b;
            else
                ypred = obj.alpha'*f+obj.b;
            end
            ypredmat=reshape(ypred,na,nb);
            
            hold off;
            plot(obj.SD.trainvec(1,obj.SD.trainlab>0),obj.SD.trainvec(2,obj.SD.trainlab>0),'k+');
            hold on;
            plot(obj.SD.trainvec(1,obj.SD.trainlab<0),obj.SD.trainvec(2,obj.SD.trainlab<0),'g+');
            plot(obj.SV(1,obj.alpha<0),obj.SV(2,obj.alpha<0),'og','MarkerSize',12,'LineWidth',2);
            plot(obj.SV(1,obj.alpha>0),obj.SV(2,obj.alpha>0),'ok','MarkerSize',12,'LineWidth',2);
            plot(obj.SV(1,1:length(obj.Iw)),obj.SV(2,1:length(obj.Iw)),'ob','MarkerSize',8,'LineWidth',1);
            plot(obj.SV(1,length(obj.Iw)+1:length(obj.Iw)+length(obj.Iup)),obj.SV(2,length(obj.Iw)+1:length(obj.Iw)+length(obj.Iup)),'or','MarkerSize',10,'LineWidth',2);
            plot(obj.SV(1,end-length(obj.Iun)+1:end),obj.SV(2,end-length(obj.Iun)+1:end),'om','MarkerSize',10,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[-1 0 1],'g');
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        
        
        function print2DMulti(obj,SMt)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            assert(nargin==2);
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            if length(SMt)==length(unique(obj.SD.trainlab));
                ypred = obj.test1vsAll(SMt,xtest);
            else
                ypred = obj.test1vs1(SMt,xtest);
            end
            ypredmat=reshape(ypred,na,nb);
            %labs = unique(obj.SD.trainlab);
            hold on;
            [c,h] = contour(xtesta1,xtesta2,ypredmat);
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        
        function y = evaluateSolution(obj,candidates)
            indd = obj.SD.getKernelValues([obj.Iw,obj.Iup,obj.Iun],candidates);
            [val,inds]= sort(indd);
            temp = [obj.alpha;obj.Slacks(obj.Iup);-obj.Slacks(obj.Iun)];
            y = obj.SD.kernelCache(candidates,val)*temp(inds)+obj.SD.trainlab(candidates)*obj.b;
        end
        
        function activateBoxConstraints(obj)
            if (~isempty(find(obj.alpha<-obj.Slacks(obj.Iw), 1)) || ...
                    ~isempty(find(obj.alpha>obj.Slacks(obj.Iw), 1))) && length(obj.alpha)>2
                dir = obj.alpha - obj.oldalpha;
                indad = find(obj.alpha < -obj.Slacks(obj.Iw));
                indsup = find(obj.alpha > obj.Slacks(obj.Iw));
                [tI indmin] = min((-obj.Slacks(obj.Iw(indad))-obj.oldalpha(indad))./dir(indad));
                [tS indS] = min((obj.Slacks(obj.Iw(indsup))-obj.oldalpha(indsup))./dir(indsup));
                if isempty(tI) , tI = tS + 1; end;
                if isempty(tS) , tS = tI + 1; end;
                t = min(tI,tS);
                obj.oldalpha = obj.oldalpha + t*dir;
                if t==tI
                    obj.Iun = sort([obj.Iun , obj.Iw(indad(indmin))]);
                    obj.Iw(indad(indmin)) = [];
                    obj.oldalpha(indad(indmin)) = [];
                    obj.alpha(indad(indmin)) = [];
                else
                    obj.Iup = sort([obj.Iup , obj.Iw(indsup(indS))]);
                    obj.Iw(indsup(indS))= [];
                    obj.oldalpha(indsup(indS))= [];
                    obj.alpha(indsup(indS)) = [];
                end
                obj.activated = 1;
            else
                obj.activated = 0;
            end
            
        end
        
         function relaxBoxConstraints(obj)
            xt = zeros(obj.SD.n,1);
            xt(obj.Iw) = obj.alpha;
            xt(obj.Iup) = obj.Slacks(obj.Iup);
            xt(obj.Iun) = -obj.Slacks(obj.Iun);
            S = 500;
            Found=0;
            nbLoop = ceil(length(obj.Il)/S);
            z = 0;
            if not(isempty(obj.Il))
                while (z<nbLoop && Found==0)
                    z = z+1;
                    nul = obj.evaluateSolution(obj.Il((obj.iloop-1)*S+1:min(obj.iloop*S,end)))-1;
                    [mm mpos]=min(nul);
                    if (mm<-obj.kktgap)
                        Found=1;
                        mpos = mpos+(obj.iloop-1)*S;
                    else
                        mm = inf; mpos = [];
                    end
                    %obj.iloop= obj.iloop+1;
                    obj.iloop = mod(obj.iloop,nbLoop)+1;
                    %keyboard
                end
            else
                mm = inf; mpos = [];
            end
            if not(isempty(obj.Iup))
                nuu = obj.evaluateSolution(obj.Iup)-1;
                [mmS mposS]=min(-nuu);
            else
                mmS = inf; mposS = [];
            end
            
            if not(isempty(obj.Iun))
                nuun = obj.evaluateSolution(obj.Iun)-1;
                [mmSn mposSn]=min(-nuun);
            else
                mmSn = inf; mposSn = [];
            end
            
            if (~isempty(mmS) && mmS < -obj.kktgap) || (~isempty(mm) && mm < -obj.kktgap)...
                    || (~isempty(mmSn) && mmSn < -obj.kktgap)
                if mm == min([mm,mmS,mmSn])
                    obj.Iw = sort([obj.Iw , obj.Il(mpos)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.Il(mpos) = [];
                elseif mmS == min([mmS,mmSn])
                    obj.Iw = sort([obj.Iw , obj.Iup(mposS)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.Iup(mposS)=[];
                else
                    obj.Iw = sort([obj.Iw , obj.Iun(mposSn)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.Iun(mposSn)=[];                    
                end
            else
                obj.converged = 1;
            end
        end
        
        function finish(obj)
            if ~isempty(obj.SD.trainvec)
                obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iup,obj.Iun]);
                
            else
                %disp('no training data can be stored for later use :
                %pre-computed matrix');
            end
            obj.alpha = [obj.alpha;obj.Slacks(obj.Iup);-obj.Slacks(obj.Iun)].*obj.SD.trainlab([obj.Iw,obj.Iup,obj.Iun]);
        end
        
        function [ypred perf] = test(obj,testdata,testlabel)
            if isempty(obj.SD.trainvec) % pre-computed case
                assert(size(testdata,1)==obj.SD.n,'simpleModel:wrongInput','precomputed-kernel used, need to provide a matrix nntrain*nbtest to compute a test error');
                ypred = (obj.alpha'*testdata([obj.Iw,obj.Iup,obj.Iun],:)+obj.b)';
                if nargin==3
                    perf = 100*length(find(sign(ypred)==testlabel))/length(testlabel);
                else
                    perf = -1;
                end
            else
                ypred = (obj.alpha'*kernelEval(obj,testdata)+obj.b)';
                if nargin==3
                    perf = 100*length(find(sign(ypred)==testlabel))/length(testlabel);
                else
                    perf = -1;
                end
            end
        end
        
        function grad = computeGradient(obj)
            grad = mean(obj.SD.kernelCache(1:obj.SD.n,[obj.Iw,obj.Iup,obj.Iun])*...
                [obj.alpha;obj.Slacks(obj.Iup);-obj.Slacks(obj.Iun)]-1);
        end
        
        function train(obj)
            figure;
            iter = 0;
            while obj.converged==0 && iter<1000*length(obj.SD.n)
                iter = iter+1;
                obj.solveLinearSystem();
                grad(iter) = obj.computeGradient();
                subplot(2,1,2);
                obj.print2D(); 
                subplot(2,1,1);
                hold off;
                plot(grad);
                obj.activateBoxConstraints();
                if (obj.activated==0)
                    obj.relaxBoxConstraints();
                end
            end
            if obj.converged==0
                disp('No convergence');
                obj.solveLinearSystem();
                obj.converged = 1;
            end
            obj.finish();
            %obj.print2D();
            
        end
        
        
    end
    
end
