classdef simpleModelDocSVDD < simpleModel
    % SIMPLEMODELDocSVM This class inherits from simpleModel. It contains
    % particular methods to solve multiclass-DOC-SVM.
    % This class support Doc-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 16/02/2011
    % Last Modified 16/12/2013 by G.L.
    
    properties
        AA;
        Q;
        lambda;
        A;
        Z;
        radius;
        centers;
    end
    
    methods
        % contructor
        function SM = simpleModelDocSVDD(SD,C,Q,lambda,Il,Iu,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Q : Q value for Doc
            %           lambda : for Doc kernel
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelDocSVM object
            
            assert(nargin==4 || nargin==7,'simpleModelDocSVDD:wrongInput','Requires 4 or 7 input arguments');
            if nargin<4
                error('not enough input');
            else
                SM.SD = SD;
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                SM.Q = Q;
                SM.lambda = lambda;
                if nargin==7
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [];
                    for q=1:Q
                        SM.Iw = [SM.Iw, find(SD.trainlab==q, 1 )];
                    end
                    SM.Iu = [];
                    SM.Il = 1:SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(2,1);
                end
                SM.labs = sort(unique(SD.trainlab));
                if length(C)==length(SD.trainlab)
                    SM.Slacks = C;
                elseif length(C)==length(SM.labs)
                    for i=1:length(SM.labs)
                        SM.Slacks(SD.trainlab==SM.labs(i)) = C(i);
                    end
                else
                    SM.Slacks = C(1)*ones(size(SD.trainlab));
                    
                end
                SM.activated = 0;
                SM.converged = 0;
                
                SM.AA = zeros(size(SD.trainlab,1),SM.Q);
                for q = 1:SM.Q
                    SM.AA(:,q) = SD.trainlab==q;
                end
                
                
                SM.A = (SM.Q-2)*SM.lambda + 1;
                SM.Z = 1+(SM.Q-2)*SM.lambda-(Q-1)*(SM.lambda^2);
            end
        end
        
        
        
        function set.Q(obj,val)
            obj.Q = val;
        end
        
        function Q = get.Q(obj)
            Q = obj.Q;
        end
        
        
        
        function set.AA(obj,val)
            obj.AA = val;
        end
        
        function AA = get.AA(obj)
            AA = obj.AA;
        end
        
        
        
        
        function set.lambda(obj,val)
            obj.lambda = val;
        end
        
        function lambda = get.lambda(obj)
            lambda = obj.lambda;
        end
        
        function set.A(obj,val)
            obj.A = val;
        end
        
        function A = get.A(obj)
            A = obj.A;
        end
        
        function set.Z(obj,val)
            obj.Z = val;
        end
        
        function Z = get.Z(obj)
            Z = obj.Z;
        end
        
        function set.radius(obj,val)
            obj.radius = val;
        end
        
        function radius = get.radius(obj)
            radius = obj.radius;
        end
        
        function set.centers(obj,val)
            obj.centers = val;
        end
        
        function centers = get.centers(obj)
            centers = obj.centers;
        end
        
        
        
        
        function activateBoxConstraints(obj)
            if (~isempty(find(obj.alpha<0, 1)) || ~isempty(find(obj.alpha>obj.Slacks(obj.Iw), 1)))...
                    && isempty(find(sum(obj.AA(obj.Iw,:))-1<=0, 1))
                dir = obj.alpha - obj.oldalpha;
                indad = find(obj.alpha < 0);
                indsup = find(obj.alpha > obj.Slacks(obj.Iw));
                [tI indmin] = min(-obj.oldalpha(indad)./dir(indad));
                [tS indS] = min((obj.Slacks(obj.Iw(indsup))-obj.oldalpha(indsup))./dir(indsup));
                if isempty(tI) , tI = tS + 1; end;
                if isempty(tS) , tS = tI + 1; end;
                t = min(tI,tS);
                obj.oldalpha = obj.oldalpha + t*dir;
                if t==tI
                    obj.Il = sort([obj.Il , obj.Iw(indad(indmin))]);
                    obj.Iw(indad(indmin)) = [];
                    obj.oldalpha(indad(indmin)) = [];
                    obj.alpha(indad(indmin)) = [];
                else
                    obj.Iu = sort([obj.Iu , obj.Iw(indsup(indS))]);
                    obj.Iw(indsup(indS))= [];
                    obj.oldalpha(indsup(indS))= [];
                    obj.alpha(indsup(indS)) = [];
                end
                obj.activated = 1;
            else
                obj.activated = 0;
            end
            
        end
        
        
        function [pred] = evaluateSolution(obj,candidates)
             %pred = zeros(length(candidates),obj.Q);
   %         for q = 1:obj.Q
%                 
%                 svq = obj.Iw(find(obj.SD.trainlab(obj.Iw)==q,1));
%                 indd = obj.SD.getDoc2KernelValues([candidates,svq],[],obj.Q,obj.lambda,obj.A,obj.Z);
%                 R2 = ((obj.A/obj.Z))*(obj.AA([obj.Iw,obj.Iu],q));
%                 R2 = R2 + ((-obj.lambda/obj.Z))*(1-obj.AA([obj.Iw,obj.Iu],q));
%                 f =  obj.SD.kernelCache( [obj.Iw,obj.Iu] ,indd)'; keyboard
%                 xtaq = f*(([obj.alpha;obj.Slacks(obj.Iu)]./R2));
%                 obj.radius(q) =  -2*xtaq(end);% +kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,svq), ...
%                   %  1, obj.SD.trainvec(:,svq), 1); %+ aqtaq
%                  
%                  pred(:,q) = obj.radius(q)  + 2*xtaq(1:end-1)  ;%diag(kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,candidates), ...
% %                     %ones(size(candidates))', obj.SD.trainvec(:,candidates), ones(size(candidates))')); %- aqtaq
% % 

                            
                indd = obj.SD.getDoc2KernelValues([candidates],[],obj.Q,obj.lambda,obj.A,obj.Z);
                %R2 = ((obj.A/obj.Z))*(obj.AA([obj.Iw,obj.Iu],q));
                %R2 = R2 +
                %((-obj.lambda/obj.Z))*(1-obj.AA([obj.Iw,obj.Iu],q));
                pred = obj.SD.kernelCache( [obj.Iw,obj.Iu] ,indd)'*(([obj.alpha;obj.Slacks(obj.Iu)]))-...
                    obj.AA(candidates,:)*obj.b -diag(obj.SD.kernelCachePur( candidates ,indd))/2 ; 
               


  %          end
            
        end
        
        function relaxBoxConstraints(obj)
            xt = zeros(obj.SD.n,1);
            xt(obj.Iw) = obj.alpha;
            xt(obj.Iu) = obj.Slacks(obj.Iu);
            %indd = obj.SD.getDoc2KernelValues([obj.Iw,obj.Iu],[],obj.Q,obj.lambda,obj.A,obj.Z);
            %[val,inds]= sort(indd);
            %temp = [obj.Iw,obj.Iu];
            S = 200;
            Found=0;
            nbLoop = ceil((length(obj.Il)-1)/S);
            z = 0;
            if not(isempty(obj.Il))
                while (z<nbLoop && Found==0)
                    z = z+1;
                    %nul = obj.SD.kernelCache(obj.Il((obj.iloop-1)*S+1:min(obj.iloop*S,end)),indd)*[obj.alpha;obj.Slacks(obj.Iu)]-1+...
                    %    obj.AA(obj.Il((obj.iloop-1)*S+1:min(obj.iloop*S,end)),:)*obj.b;
                    [nul] = obj.evaluateSolution(obj.Il((obj.iloop-1)*S+1:min(obj.iloop*S,end))); 
                    %%nul= max(nul,[],2)%.*obj.AA(obj.Il((obj.iloop-1)*S+1:min(obj.iloop*S,end)),:),2);
                    %nul=sum(nul.*obj.AA(obj.Il((obj.iloop-1)*S+1:min(obj.iloop*S,end)),:),2);
                    [mm mpos]=min(nul);
                    if (mm<-obj.kktgap)
                        Found=1;
                        mpos = mpos+(obj.iloop-1)*S;
                    else
                        mm = inf; mpos = [];
                    end
                    %obj.iloop= obj.iloop+1;
                    obj.iloop = mod(obj.iloop,nbLoop)+1;
                    %keyboard
                end
            else
                mm = inf; mpos = [];
            end
            if not(isempty(obj.Iu))
                %fx = obj.SD.getDocKernelValues([obj.Iw,obj.Iu],obj.Iu,obj.Q,obj.lambda);
                %nuu = obj.SD.kernelCache(obj.Iu,indd)*[obj.alpha;obj.Slacks(obj.Iu)]-1+...
                %        obj.AA(obj.Iu,:)*obj.b;
                [nuu] = obj.evaluateSolution(obj.Iu);
                %nuu = sum(nuu.*obj.AA(obj.Iu,:),2);
                [mmS mposS]=min(-nuu);
            else
                mmS = inf; mposS = [];
            end
            if (~isempty(mmS) && mmS < -obj.kktgap) || (~isempty(mm) && mm < -obj.kktgap)
                if isempty(mmS) || mm < mmS
                    obj.Iw = sort([obj.Iw , obj.Il(mpos)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.Il(mpos) = [];
                else
                    obj.Iw = sort([obj.Iw , obj.Iu(mposS)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.Iu(mposS)=[];
                end
            else
                obj.converged = 1;
            end
        end
        
        
        function solveLinearSystem(obj)
            be = 1;
            indd = obj.SD.getDoc2KernelValues(obj.Iw,obj.Iw,obj.Q,obj.lambda, obj.A,obj.Z);
            ce =  0.5*diag(obj.SD.kernelCachePur(obj.Iw,indd))';
            if ~isempty(obj.Iu)
                
                ce = ce - (obj.Slacks(obj.Iu))'*obj.SD.kernelCache(obj.Iu,indd);
                be = be - obj.AA(obj.Iu,:)'*obj.Slacks(obj.Iu);
            end
            if obj.optimType==1
                G = chol(obj.SD.kernelCache(obj.Iw,indd)+obj.kktgap*eye(length(obj.Iw))); 
                M = obj.AA(obj.Iw,:)'*(G\(G'\obj.AA(obj.Iw,:)));
                obj.b = ((M)\((obj.AA(obj.Iw,:)'*(G\(G'\-ce'))+be)));
                obj.alpha = (G\(G'\(ce'+obj.AA(obj.Iw,:)*obj.b)));
            else
                error('not implemented yet');
            end
            %obj.alpha = obj.alpha*((obj.A+(obj.Q-1)*obj.lambda)/obj.Z)^2;
        end
        
        
        function reinit(obj)
            obj.Iw = [];
            for q=1:obj.Q
                obj.Iw = [obj.Iw, find(obj.SD.trainlab==q, 1 )];
            end
            obj.Iu = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            
            if length(obj.Slacks)~=length(obj.SD.trainlab)
                if length(obj.C)==length(obj.SD.trainlab)
                    obj.Slacks = obj.C;
                elseif length(obj.C)==length(obj.labs)
                    for i=1:length(obj.labs)
                        obj.Slacks(obj.SD.trainlab==obj.labs(i)) = obj.C(i);
                    end
                else
                    obj.Slacks = obj.C(1)*ones(size(obj.SD.trainlab));
                    
                end
            end
            obj.activated = 0;
            obj.converged = 0;
            
            obj.AA = zeros(size(obj.SD.trainlab,1),obj.Q);
            for q = 1:obj.Q
                obj.AA(:,q) = obj.SD.trainlab==q;
            end
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelDocSVM:print2D','must be 2D data');
            scale = 3;
            [xtesta1,xtesta2]=meshgrid([-scale:0.01:scale],[-scale:0.01:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            [ypred,pred] = obj.test(xtest,ones(length(xtest),1));
            color = ['bgymckbgymcbgymcbgymc'];
            hold off;
            ypredmat=reshape(pred,na,nb);
            [c,h] = contourf(xtesta1,xtesta2,ypredmat); colormap('gray');%colorbar
            hold on;
            for q=1:obj.Q
                plot(obj.SD.trainvec(1,obj.SD.trainlab==q),obj.SD.trainvec(2,obj.SD.trainlab==q),[color(q),'.'],'linewidth',3);
                ypredmat=reshape(ypred(:,q),na,nb);
                %hold on;
                [c,h] = contour(xtesta1,xtesta2,ypredmat,[0 0],'.-');
                %[c,h2] = contour(xtesta1,xtesta2,ypredmat,[-obj.radius(q) obj.radius(q)]);
                %
                set(h,'LineWidth',1);
                set(h,'color',color(q));
                %set(h2,'color','w');
            end
            %[~,ypred] = max(ypred,[],2);
            %set(h,'LineWidth',2);
            plot(obj.SD.trainvec(1,obj.Iw),obj.SD.trainvec(2,obj.Iw),'ko','markersize',10);
            plot(obj.SD.trainvec(1,obj.Iu),obj.SD.trainvec(2,obj.Iu),'ks','markersize',10);
            
            drawnow;
            
        end
        
        
        
        function [pred ypred perf] = test(obj,testdata,testlabel)
            if isempty(obj.SD.trainvec) % pre-computed case
                %                 assert(size(testdata,1)==obj.SD.n,'simpleModel:wrongInput','precomputed-kernel used, need to provide a matrix nntrain*nbtest to compute a test error');
                %                 ypred = (obj.alpha'*testdata([obj.Iw,obj.Iu],:)+obj.b)';
                %                 if nargin==3
                %                     perf = 100*length(find(sign(ypred)==testlabel))/length(testlabel);
                %                 else
                %                     perf = -1;
                %                 end
            else
                pred = zeros(size(testdata,2),obj.Q);
                if obj.converged==1
                    for q = 1:obj.Q
                        R = (obj.A/obj.Z)*(obj.AA([obj.Iw,obj.Iu],q));
                        R = R + (-obj.lambda/obj.Z)*(1-obj.AA([obj.Iw,obj.Iu],q));
                        %R = R*R'
                        f =  kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ...
                            ones(size([obj.Iw,obj.Iu]))',testdata, ones(size(testlabel)));
                        xtaq = f*(obj.alpha.*R);
                        pred(:,q) = obj.radius(q) + 2*xtaq  - obj.centers(q) -1 ;%  - diag(kernelTest(obj.SD.kernelType,obj.SD.kernelParam,testdata, ...
                           %testlabel, testdata, testlabel)); %- aqtaq

                    end
                    %pred = pred-(ones(length(pred),1)*(obj.centers));
                else
                    for q = 1:obj.Q
                        pred(:,q)=kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ...
                            ones(size(obj.SD.trainlab([obj.Iw,obj.Iu]))), testdata,...
                            q,obj.Q, obj.lambda)*[obj.alpha;obj.Slacks(obj.Iu)];
                    end
                end
                
                [~,ypred] = max(pred,[],2);
                if nargin==3
                    perf = 100*length(find(ypred==testlabel))/length(testlabel);
                    %keyboard
                else
                    perf = -1;
                end
            end
        end
        
        
        
        function finish(obj)
            if ~isempty(obj.SD.trainvec)
                obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
                
            else
                %disp('no training data can be stored for later use :
                %pre-computed matrix');
            end
            obj.alpha = [obj.alpha;obj.Slacks(obj.Iu)];
            
            for q=1:obj.Q
                %T = ((obj.A/obj.Z)^2)*(obj.AA([obj.Iw,obj.Iu],q)*obj.AA([obj.Iw,obj.Iu],q)');
                %T = T + ((obj.lambda/obj.Z)^2)*(1-obj.AA([obj.Iw,obj.Iu],q))*(1-obj.AA([obj.Iw,obj.Iu],q))';
                %T = T + ((-obj.A*obj.lambda)/obj.Z^2)*((1-obj.AA([obj.Iw,obj.Iu],q))*obj.AA([obj.Iw,obj.Iu],q)');
                %T = T + ((-obj.A*obj.lambda)/obj.Z^2)*(obj.AA([obj.Iw,obj.Iu],q)*(1-obj.AA([obj.Iw,obj.Iu],q))');
                
                %f =  kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ...
                %    ones(size([obj.Iw,obj.Iu]))', obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ones(size([obj.Iw,obj.Iu]))');
                
                indd = obj.SD.getDoc2KernelValues([obj.Iw,obj.Iu],[],obj.Q,obj.lambda,obj.A,obj.Z);
                f = obj.SD.kernelCachePur( [obj.Iw,obj.Iu] ,indd);
                
                R2 = ((obj.A/obj.Z))*(obj.AA([obj.Iw,obj.Iu],q));
                R2 = R2 + ((-obj.lambda/obj.Z))*(1-obj.AA([obj.Iw,obj.Iu],q));
                
                %norm(T-R2*R2')
                %keyboard
                
                %indd = obj.SD.getDoc2KernelValues([obj.Iw,obj.Iu],[],obj.Q,obj.lambda);
                %f =  obj.SD.kernelCache( [obj.Iw,obj.Iu] ,indd)';
                obj.centers(q) = (obj.alpha.*R2)'*(f)*(obj.alpha.*R2);
                %obj.centers(q) = (obj.alpha)'*(f.*T)*(obj.alpha);
                ii = find(obj.SD.trainlab(obj.Iw)==q,1);
                %svq = obj.Iw(ii);
                %indd = obj.SD.getDoc2KernelValues(svq,[],obj.Q,obj.lambda);
                %f =  obj.SD.kernelCache( [obj.Iw,obj.Iu] ,indd)';
                xtaq2 = f(ii,:)*(obj.alpha.*R2);
                
                obj.radius(q) = obj.centers(q) - 2*xtaq2 +f(ii,ii); %+ aqtaq
            end
            
            %obj.radius = obj.radius+obj.centers;
        end
        
        function err = crossValError(obj,nbFold)
            assert(nbFold<=obj.SD.n && nbFold>0,'simpleModel:wrongIput','nbFold must range between 1 and number of training points');
            Sl = obj.Slacks;
            SDSaved = obj.SD;
            if ~isempty(obj.SD.trainvec)
                s = ceil(obj.SD.n/nbFold);
                p = zeros(1,nbFold);
                rn = randperm(SDSaved.n);
                for i=1:nbFold
                    il = rn((i-1)*s+1:min(i*s,SDSaved.n));
                    it = setdiff(1:SDSaved.n,il);
                    SDt = simpleData(SDSaved.trainvec(:,it),SDSaved.trainlab(it));
                    obj.SD = SDt;
                    obj.Slacks = Sl(it);
                    obj.clear();
                    obj.reinit();
                    obj.train();
                    [~ , p(i)] = obj.test(SDSaved.trainvec(:,il),SDSaved.trainlab(il));
                    clear SDt;
                end
            else
                %                 SDSaved.kernelCache = SDSaved.kernelCache.*(SDSaved.trainlab*SDSaved.trainlab');
                %
                %                 s = ceil(obj.SD.n/nbFold);
                %                 p = zeros(1,nbFold);
                %                 rn = randperm(SDSaved.n);
                %                 for i=1:nbFold
                %                     il = rn((i-1)*s+1:min(i*s,SDSaved.n));
                %                     it = setdiff(1:SDSaved.n,il);
                %                     SDt = simpleData(SDSaved.kernelCache(it,it),SDSaved.trainlab(it));
                %                     obj.SD = SDt;
                %                     obj.Slacks = Sl(it);
                %                     obj.clear();
                %                     obj.reinit();
                %                     obj.train();
                %                     [~ , p(i)] = obj.test(SDSaved.kernelCache(it,il),SDSaved.trainlab(il));
                %                     clear SDt;
                %end
               % SDSaved.kernelCache = SDSaved.kernelCache.*(SDSaved.trainlab*SDSaved.trainlab');
            end
            err = 100-mean(p);
            obj.SD = SDSaved();
            obj.Slacks = Sl;
        end
        
        
        function clear(obj)
            obj.alpha = [];
            obj.SV = [];
            obj.b = [];
            obj.iloop = 1;
            obj.AA = [];
        end
        
    end
    
end
