classdef simpleModelDocSVM < simpleModel
    % SIMPLEMODELDocSVM This class inherits from simpleModel. It contains
    % particular methods to solve multiclass-DOC-SVM.
    % This class support Doc-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 16/02/2011
    % Last Modified 16/02/2011 by G.L.
    
    properties
        AA;
        Q;
        lambda;
    end
    
    methods
        % contructor
        function SM = simpleModelDocSVM(SD,C,Q,lambda,Il,Iu,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Q : Q value for Doc
            %           lambda : for Doc kernel
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelDocSVM object
            
            assert(nargin==4 || nargin==7,'simpleModelCSVM:wrongInput','Requires 4 or 7 input arguments');
            if nargin<4
                error('not enough input');
            else
                SM.SD = SD;
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                SM.Q = Q;
                SM.lambda = lambda;
                if nargin==7
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [];
                    for q=1:Q
                        SM.Iw = [SM.Iw, find(SD.trainlab==q, 1 )];
                    end
                    SM.Iu = [];
                    SM.Il = 1:SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(2,1);
                end
                SM.labs = sort(unique(SD.trainlab));
                if length(C)==length(SD.trainlab)
                    SM.Slacks = C;
                elseif length(C)==length(SM.labs)
                    for i=1:length(SM.labs)
                        SM.Slacks(SD.trainlab==SM.labs(i)) = C(i);
                    end
                else
                    SM.Slacks = C(1)*ones(size(SD.trainlab));
                    
                end
                SM.activated = 0;
                SM.converged = 0;
                
                SM.AA = zeros(size(SD.trainlab,1),SM.Q);
                for q = 1:SM.Q
                    SM.AA(:,q) = SD.trainlab==q;
                end
            end
        end
        
        
        
        function set.Q(obj,val)
            obj.Q = val;
        end
        
        function Q = get.Q(obj)
            Q = obj.Q;
        end
        
        
        
        function set.AA(obj,val)
            obj.AA = val;
        end
        
        function AA = get.AA(obj)
            AA = obj.AA;
        end
        
        
        
        
        function set.lambda(obj,val)
            obj.lambda = val;
        end
        
        function lambda = get.lambda(obj)
            lambda = obj.lambda;
        end
        
        
        
        
        function activateBoxConstraints(obj)
            if (~isempty(find(obj.alpha<0, 1)) || ~isempty(find(obj.alpha>obj.Slacks(obj.Iw), 1)))...
                    && isempty(find(sum(obj.AA(obj.Iw,:))-1<=0, 1))
                dir = obj.alpha - obj.oldalpha;
                indad = find(obj.alpha < 0);
                indsup = find(obj.alpha > obj.Slacks(obj.Iw));
                [tI indmin] = min(-obj.oldalpha(indad)./dir(indad));
                [tS indS] = min((obj.Slacks(obj.Iw(indsup))-obj.oldalpha(indsup))./dir(indsup));
                if isempty(tI) , tI = tS + 1; end;
                if isempty(tS) , tS = tI + 1; end;
                t = min(tI,tS);
                obj.oldalpha = obj.oldalpha + t*dir;
                if t==tI
                    obj.Il = sort([obj.Il , obj.Iw(indad(indmin))]);
                    obj.Iw(indad(indmin)) = [];
                    obj.oldalpha(indad(indmin)) = [];
                    obj.alpha(indad(indmin)) = [];
                else
                    obj.Iu = sort([obj.Iu , obj.Iw(indsup(indS))]);
                    obj.Iw(indsup(indS))= [];
                    obj.oldalpha(indsup(indS))= [];
                    obj.alpha(indsup(indS)) = [];
                end
                obj.activated = 1;
            else
                obj.activated = 0;
            end
            
        end
        
        
        function y = evaluateSolution(obj,candidates)
            pred = zeros(length(candidates),obj.Q);
            for q = 1:obj.Q
                pred(:,q)=kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ...
                    obj.SD.trainlab([obj.Iw,obj.Iu]), obj.SD.trainvec(:,candidates),...
                    q, obj.lambda)*[obj.alpha;obj.Slacks(obj.Iu)]-obj.b(q);
            end
            [~,y] = max(pred,[],2);
            %y = obj.SD.getKernelValues([obj.Iw,obj.Iu],candidates)*[obj.alpha;obj.Slacks(obj.Iu)]+obj.AA(candidates,:)*obj.b;
        end
        
        function relaxBoxConstraints(obj)
            xt = zeros(obj.SD.n,1);
            xt(obj.Iw) = obj.alpha;
            xt(obj.Iu) = obj.Slacks(obj.Iu);
            indd = obj.SD.getDocKernelValues([obj.Iw,obj.Iu],obj.Il,obj.Q,obj.lambda);
            [val,inds]= sort(indd);
            temp = [obj.Iw,obj.Iu];
            S = 500;%round(1000*rand(1));
            Found=0;
            nbLoop = ceil(length(obj.Il)/S);
            z = 0;
            if not(isempty(obj.Il))
                while (z<nbLoop && Found==0)
                    z = z+1;
                    nul = obj.SD.kernelCache(obj.Il((obj.iloop-1)*S+1:min(obj.iloop*S,end)),val)*xt(temp(inds)) - ...
                        obj.AA(obj.Il((obj.iloop-1)*S+1:min(obj.iloop*S,end)),:)*obj.b;
                    [mm mpos]=min(nul);
                    if (mm<-obj.kktgap)
                        Found=1;
                        mpos = mpos+(obj.iloop-1)*S;
                    else
                        mm = inf; mpos = [];
                    end
                    %obj.iloop= obj.iloop+1;
                    obj.iloop = mod(obj.iloop,nbLoop)+1;
                    %keyboard
                end
            else
                mm = inf; mpos = [];
            end
            if not(isempty(obj.Iu))
                %fx = obj.SD.getDocKernelValues([obj.Iw,obj.Iu],obj.Iu,obj.Q,obj.lambda);
                nuu = obj.SD.kernelCache(obj.Iu,val)*xt(temp(inds)) - ...
                    obj.AA(obj.Iu,:)*obj.b;
                [mmS mposS]=min(-nuu);
            else
                mmS = inf; mposS = [];
            end
            
            if (~isempty(mmS) && mmS < -obj.kktgap) || (~isempty(mm) && mm < -obj.kktgap)
                if isempty(mmS) || mm < mmS
                    obj.Iw = sort([obj.Iw , obj.Il(mpos)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.Il(mpos) = [];
                else
                    obj.Iw = sort([obj.Iw , obj.Iu(mposS)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.Iu(mposS)=[];
                end
            else
                obj.converged = 1;
            end
        end
        
        
        function solveLinearSystem(obj)
            ce = zeros(1,length(obj.Iw));
            be = 1;
            indd = obj.SD.getDocKernelValues(obj.Iw,obj.Iw,obj.Q,obj.lambda);
            if ~isempty(obj.Iu)
                
                ce = ce + (obj.Slacks(obj.Iu))'*obj.SD.kernelCache(obj.Iu,indd);
                be = be - obj.AA(obj.Iu,:)'*obj.Slacks(obj.Iu);
            end
            if obj.optimType==1
                %G = chol(obj.SD.getDocKernelValues(obj.Iw,obj.Iw,obj.Q,obj.lambda));
                G = chol(obj.SD.kernelCache(obj.Iw,indd)+obj.kktgap*eye(length(obj.Iw)));
                M = obj.AA(obj.Iw,:)'*(G\(G'\obj.AA(obj.Iw,:)));
                obj.b = (M)\(obj.AA(obj.Iw,:)'*(G\(G'\ce'))+be);
                obj.alpha = G\(G'\(-ce'+obj.AA(obj.Iw,:)*obj.b));
            else
                error('not implemented yet');
            end
            
            
            
            %             A = svModel.temp.AA(svModel.temp.active,:);
            %             B = svModel.temp.kerC(svModel.temp.active);
            %             [~, b] = sort(svModel.temp.inverse(1:svModel.temp.n_active));
            %             A = A(b,:);
            %             B = B(b);
            %             M = A'*svModel.model.G1(svModel.temp.n_H,svModel.temp.n_H);
            %             svModel.model.mu = (M*A)\(1-svModel.temp.cardIc+M*B);
            %             svModel.model.alpha = svModel.model.G1(svModel.temp.n_H,svModel.temp.n_H)*(A*...
            %                 svModel.model.mu-B);
        end
        
        
        function reinit(obj)
            obj.Iw = [];
            for q=1:obj.Q
                obj.Iw = [obj.Iw, find(obj.SD.trainlab==q, 1 )];
            end
            obj.Iu = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            
            if length(obj.Slacks)~=length(obj.SD.trainlab)
                if length(obj.C)==length(obj.SD.trainlab)
                    obj.Slacks = obj.C;
                elseif length(obj.C)==length(obj.labs)
                    for i=1:length(obj.labs)
                        obj.Slacks(obj.SD.trainlab==obj.labs(i)) = obj.C(i);
                    end
                else
                    obj.Slacks = obj.C(1)*ones(size(obj.SD.trainlab));
                    
                end
            end
            obj.activated = 0;
            obj.converged = 0;
            
            obj.AA = zeros(size(obj.SD.trainlab,1),obj.Q);
            for q = 1:obj.Q
                obj.AA(:,q) = obj.SD.trainlab==q;
            end
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelDocSVM:print2D','must be 2D data');
            scale = 3;
            [xtesta1,xtesta2]=meshgrid([-scale:0.01:scale],[-scale:0.01:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            ypred = obj.test(xtest);
            color = ['bgymckbgymcbgymcbgymc'];
            hold off;
            [~,ypred2] = max(ypred,[],2);
            ypredmat=reshape(ypred2,na,nb);
            hold on;
            [c,h] = contourf(xtesta1,xtesta2,ypredmat);colormap('gray');
            %set(h,'LineWidth',2);
            for q=1:obj.Q
                plot(obj.SD.trainvec(1,obj.SD.trainlab==q),obj.SD.trainvec(2,obj.SD.trainlab==q),[color(q),'.'],'linewidth',3);
                ypredmat=reshape(ypred(:,q),na,nb);
                %hold on;
                [c,h] = contour(xtesta1,xtesta2,ypredmat,[0 0],color(q));
                set(h,'LineWidth',1);
            end
            plot(obj.SD.trainvec(1,obj.Iw),obj.SD.trainvec(2,obj.Iw),'ko','markersize',10);
            plot(obj.SD.trainvec(1,obj.Iu),obj.SD.trainvec(2,obj.Iu),'ks','markersize',10);
            
            drawnow;
            
        end
        
        
        
        function [pred perf] = test(obj,testdata,testlabel)
            if isempty(obj.SD.trainvec) % pre-computed case
                %                 assert(size(testdata,1)==obj.SD.n,'simpleModel:wrongInput','precomputed-kernel used, need to provide a matrix nntrain*nbtest to compute a test error');
                %                 ypred = (obj.alpha'*testdata([obj.Iw,obj.Iu],:)+obj.b)';
                %                 if nargin==3
                %                     perf = 100*length(find(sign(ypred)==testlabel))/length(testlabel);
                %                 else
                %                     perf = -1;
                %                 end
            else
                pred = zeros(size(testdata,2),obj.Q);
                if obj.converged==1
                    for q = 1:obj.Q
                        pred(:,q)=kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SV, ...
                            obj.SD.trainlab([obj.Iw,obj.Iu]), testdata,...
                            q,obj.Q, obj.lambda)*obj.alpha-obj.b(q);
                    end
                else
                    for q = 1:obj.Q
                        pred(:,q)=kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ...
                            obj.SD.trainlab([obj.Iw,obj.Iu]), testdata,...
                            q,obj.Q, obj.lambda)*[obj.alpha;obj.Slacks(obj.Iu)]-obj.b(q);
                    end
                end
                
                [~,ypred] = max(pred,[],2);
                if nargin==3
                    perf = 100*length(find(ypred==testlabel))/length(testlabel);
                else
                    perf = -1;
                end
            end
        end
        
        
        
        function finish(obj)
            if ~isempty(obj.SD.trainvec)
                obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
                
            else
                %disp('no training data can be stored for later use :
                %pre-computed matrix');
            end
            obj.alpha = [obj.alpha;obj.Slacks(obj.Iu)];
        end
        
        function err = crossValError(obj,nbFold)
            assert(nbFold<=obj.SD.n && nbFold>0,'simpleModel:wrongIput','nbFold must range between 1 and number of training points');
            Sl = obj.Slacks;
            SDSaved = obj.SD;
            if ~isempty(obj.SD.trainvec)
                s = ceil(obj.SD.n/nbFold);
                p = zeros(1,nbFold);
                rn = randperm(SDSaved.n);
                for i=1:nbFold
                    il = rn((i-1)*s+1:min(i*s,SDSaved.n));
                    it = setdiff(1:SDSaved.n,il);
                    SDt = simpleData(SDSaved.trainvec(:,it),SDSaved.trainlab(it));
                    obj.SD = SDt;
                    obj.Slacks = Sl(it);
                    obj.clear();
                    obj.reinit();
                    obj.train();
                    [~ , p(i)] = obj.test(SDSaved.trainvec(:,il),SDSaved.trainlab(il));
                    clear SDt;
                end
            else
                %                 SDSaved.kernelCache = SDSaved.kernelCache.*(SDSaved.trainlab*SDSaved.trainlab');
                %
                %                 s = ceil(obj.SD.n/nbFold);
                %                 p = zeros(1,nbFold);
                %                 rn = randperm(SDSaved.n);
                %                 for i=1:nbFold
                %                     il = rn((i-1)*s+1:min(i*s,SDSaved.n));
                %                     it = setdiff(1:SDSaved.n,il);
                %                     SDt = simpleData(SDSaved.kernelCache(it,it),SDSaved.trainlab(it));
                %                     obj.SD = SDt;
                %                     obj.Slacks = Sl(it);
                %                     obj.clear();
                %                     obj.reinit();
                %                     obj.train();
                %                     [~ , p(i)] = obj.test(SDSaved.kernelCache(it,il),SDSaved.trainlab(il));
                %                     clear SDt;
                %end
               % SDSaved.kernelCache = SDSaved.kernelCache.*(SDSaved.trainlab*SDSaved.trainlab');
            end
            err = 100-mean(p);
            obj.SD = SDSaved();
            obj.Slacks = Sl;
        end
        
        
        function clear(obj)
            obj.alpha = [];
            obj.SV = [];
            obj.b = [];
            obj.iloop = 1;
            obj.AA = [];
        end
        
    end
    
end
