classdef simpleModelLESVM < simpleModel
    % SIMPLEMODELCSVM This class inherits from simpleModel. It contains
    % particular methods to solve binary-C-SVM.
    % This class support C-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 13/12/2011
    % Last Modified 13/12/2011 by G.L.
    
    
    
    methods
        % contructor
        function SM = simpleModelLESVM(SD,C,Il,Iu,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelCSVM object
            
            assert(nargin==2 || nargin==5,'simpleModelLESVM:wrongInput','Requires 2 or 5 input arguments');
            if nargin<2
                error('not enough input');
            else
                SM.SD = SD;
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                if nargin==5
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [find(SD.trainlab==-1,1),find(SD.trainlab==1,1)];
                    SM.Iu = [];
                    SM.Il = 1:SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(2,1);
                end
                SM.labs = sort(unique(SD.trainlab));
                if length(C)==length(SD.trainlab)
                    SM.Slacks = C;
                elseif length(C)==length(SM.labs)
                    for i=1:length(SM.labs)
                        SM.Slacks(SD.trainlab==SM.labs(i)) = C(i);
                    end
                else
                    SM.Slacks = C(1)*ones(size(SD.trainlab));
                    
                end
                SM.activated = 0;
                SM.converged = 0;
            end
        end
        
        
        
        
        
        function solveLinearSystem(obj)
            ce = ones(1,length(obj.Iw));
            be = 0;
            indd = obj.SD.getKernelValues(obj.Iw,obj.Iu);
            if ~isempty(obj.Iu)
                ce = ce - (obj.Slacks(obj.Iu))'*obj.SD.kernelCache(obj.Iu,indd);
                be = be - obj.Slacks(obj.Iu)'*obj.SD.trainlab(obj.Iu);
            end
            if obj.optimType==1
                [Q,R] = qr(obj.SD.kernelCache(obj.Iw,indd));
                M = obj.SD.trainlab(obj.Iw)'*(R\(Q'*obj.SD.trainlab(obj.Iw)));
                obj.b = M\(obj.SD.trainlab(obj.Iw)'*(R\(Q'*ce'))-be);
                obj.alpha = R\(Q'*(ce'-obj.SD.trainlab(obj.Iw)*obj.b));
            else
                error('not implemented yet');
            end
        end
        
        function train(obj)
            iter = 0;
            while obj.converged==0 && iter<1000*length(obj.SD.n)
                iter = iter+1;
                obj.solveLinearSystem();
                obj.print2D(); 
                keyboard
                obj.activateBoxConstraints();
                if (obj.activated==0)
                    obj.relaxBoxConstraints();
                end
            end
            if obj.converged==0
                disp('No convergence');
                obj.solveLinearSystem();
                obj.converged = 1;
            end
            obj.finish();
            %obj.print2D();
            
        end
        
        function [good bad] = searchSDP(obj,K)
            good = 1:size(K,1);
            bad = [];
            p=1;
            while p>0
               [~,p] = chol(K(good,good)+0.1*eye(length(good)));
               if p>0
                   i = good(p);
                   good(p) = [];
                   bad = [bad,i];
               end
            end
            if ~isempty(bad)
                [U,p] = chol(K(bad,bad)+0.1*eye(length(bad)));
                p
            end
        end
        
        function activateBoxConstraints(obj)
            % recherche des points qui empechent chol
            % cote positif
            gp = obj.Iw(obj.alpha>0);
            good=[];bad=[];good2=[];bad2=[];
            if ~isempty(gp)
                indd = obj.SD.getKernelValues(gp,gp);
                K = obj.SD.kernelCache(gp,indd);
                [good,bad] = obj.searchSDP(K);
                pp = find(obj.alpha>0);
                bad = pp(bad);
                good = pp(good);
            else
                bad = [];
            end
            gn = obj.Iw(obj.alpha<0);
            if ~isempty(gn)
                indd = obj.SD.getKernelValues(gn,gn);
                K = obj.SD.kernelCache(gn,indd);
                [good2,bad2] = obj.searchSDP(K);
                nn = find(obj.alpha<0);
                bad2 = nn(bad2);
                good2 = nn(good2);
            else
                bad2 = [];
            end
            
            if (~isempty(bad) || ~isempty(bad2)) && length(obj.alpha)>4
                dir = obj.alpha - obj.oldalpha;
                
                indad = bad;
                if ~isempty(good)
                    indsup = find(obj.alpha(good) > abs(obj.Slacks(obj.Iw(good))));
                else
                    indsup = [];
                end
                
                indadn = bad2;
                if ~isempty(good2)
                    indsupn = find(obj.alpha(good2) < -abs(obj.Slacks(obj.Iw(good2))));
                else
                    indsupn = [];
                end
                
                [tI indmin] = min(-obj.oldalpha(indad)./dir(indad));
                [tS indS] = min((abs(obj.Slacks(obj.Iw(indsup)))-obj.oldalpha(indsup))./dir(indsup));
                
                [tIn indminn] = min(-obj.oldalpha(indadn)./dir(indadn));
                [tSn indSn] = min((-abs(obj.Slacks(obj.Iw(indsupn)))-obj.oldalpha(indsupn))./dir(indsupn));
                
                if isempty(tI) || tI==0 , tI = inf; end;
                if isempty(tS) || tS==0 , tS = inf; end;
                if isempty(tIn) || tIn==0 , tIn = inf; end;
                if isempty(tSn) || tSn==0 , tSn = inf; end;
                t = min([tI,tS,tIn,tSn]);
                obj.oldalpha = obj.oldalpha + t*dir;
                t
                if t==tI
                    obj.Il = sort([obj.Il , obj.Iw(indad(indmin))]);
                    obj.Iw(indad(indmin)) = [];
                    obj.oldalpha(indad(indmin)) = [];
                    obj.alpha(indad(indmin)) = [];
                elseif t==tS
                    obj.Iu = sort([obj.Iu , obj.Iw(indsup(indS))]);
                    obj.Iw(indsup(indS))= [];
                    obj.oldalpha(indsup(indS))= [];
                    obj.alpha(indsup(indS)) = [];
                    obj.Slacks(gp(indsup(indS))) = abs(obj.Slacks(obj.Iw(indsup(indS))));
                elseif t==tIn
                    obj.Il = sort([obj.Il , obj.Iw(indadn(indminn))]);
                    obj.Iw(indad(indmin)) = [];
                    obj.oldalpha(indad(indmin)) = [];
                    obj.alpha(indad(indmin)) = [];
                elseif t==tSn
                    obj.Iu = sort([obj.Iu , obj.Iw(indsupn(indSn))]);
                    obj.Iw(indsupn(indSn))= [];
                    obj.oldalpha(indsupn(indSn))= [];
                    obj.alpha(indsupn(indSn)) = [];
                    obj.Slacks(gn(indsupn(indSn))) = -abs(obj.Slacks(obj.Iw(indsupn(indSn))));
                end
                obj.activated = 1;
            else
                obj.activated = 0;
            end
            
        end
        
        function reinit(obj)
            obj.Iw = [find(obj.SD.trainlab==-1,1),find(obj.SD.trainlab==1,1)];
            obj.Iu = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            
            if length(obj.Slacks)~=length(obj.SD.trainlab)
                if length(obj.C)==length(obj.SD.trainlab)
                    obj.Slacks = obj.C;
                elseif length(obj.C)==length(obj.labs)
                    for i=1:length(obj.labs)
                        obj.Slacks(obj.SD.trainlab==obj.labs(i)) = obj.C(i);
                    end
                else
                    obj.Slacks = obj.C(1)*ones(size(obj.SD.trainlab));
                    
                end
            end
            obj.activated = 0;
            obj.converged = 0;
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
            f = kernelEval(obj,xtest);
            if obj.converged==0
                ypred = ([obj.alpha;obj.Slacks(obj.Iu)].*obj.SD.trainlab([obj.Iw,obj.Iu]))'*f+obj.b;
            else
                ypred = obj.alpha'*f+obj.b;
            end
            ypredmat=reshape(ypred,na,nb);
            
            hold off;
            plot(obj.SD.trainvec(1,:),obj.SD.trainvec(2,:),'k+');
            hold on;
            plot(obj.SV(1,:),obj.SV(2,:),'om','MarkerSize',10,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ 0 0],'g');
            set(h,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ -1 1],'y');
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        
        
        function print2DMulti(obj,SMt)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            assert(nargin==2);
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            if length(SMt)==length(unique(obj.SD.trainlab));
                ypred = obj.test1vsAll(SMt,xtest);
            else
                ypred = obj.test1vs1(SMt,xtest);
            end
            ypredmat=reshape(ypred,na,nb);
            %labs = unique(obj.SD.trainlab);
            hold on;
            [c,h] = contour(xtesta1,xtesta2,ypredmat);
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
    end
    
end
