classdef simpleMKLModelCSVM < simpleModel
    % SIMPLEMKLMODELCSVM This class inherits from simpleModel. It contains
    % particular methods to solve Support Vector Data Description
    % This class support SVDD solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 29/11/2013
    % Last Modified 29/11/2013 by G.L.
    
    
    
    methods
        % contructor
        function SM = simpleMKLModelCSVM(SD,C,Il,Iu,Iw)
            % simpleModelSVDD : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelSVDD object
            
            assert(nargin==2 || nargin==5,'simpleModelSVDD:wrongInput','Requires 2 or 5 input arguments');
            if nargin<2
                error('not enough input');
            else
                SD.getKernelValues(1:SD.n); % pre-compute kernel
                SM.SD = SD;
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                if nargin==5
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [find(SD.trainlab==-1,1),find(SD.trainlab==1,1)];
                    SM.Iu = [];
                    SM.Il = 1:SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(2,1);
                end
                SM.dm = ones(1,length(SD.nbKernels))./length(SD.nbKernels);
                SM.labs = sort(unique(SD.trainlab));
                if length(C)==length(SD.trainlab)
                    SM.Slacks = C;
                elseif length(C)==length(SM.labs)
                    for i=1:length(SM.labs)
                        SM.Slacks(SD.trainlab==SM.labs(i)) = C(i);
                    end
                else
                    SM.Slacks = C(1)*ones(size(SD.trainlab));
                    
                end
                SM.activated = 0;
                SM.converged = 0;
            end
        end
        
        
        function train(obj)
            iter = 0;
            while obj.converged==0 && iter<1000*length(obj.SD.n)
                iter = iter+1;
                K = obj.SD.computeKernelCombination(obj.dm,1:obj.SD.n);
                SM = simpleModelCSVM(K,obj.C,obj.Il, obj.Iu, obj.Iw);
                SM.train;
                J = (SM.alpha.*obj.SD.trainlab([obj.Iw,obj.Iu]))*K([SM.Iw;SM.Iu],[SM.Iw,SM.Iu])*...
                    (SM.alpha.*obj.SD.trainlab([obj.Iw,obj.Iu]))' - sum((SM.alpha.*obj.SD.trainlab([obj.Iw,obj.Iu])));
                [~,mu] = max(obj.dm);
                for m=1:length(obj.dm)
                    dJ(m) = 0.5*SM.alpha*obj.SD.kernelCachePart(m,[SM.Iw;SM.Iu])*SM.alpha';
                end
                for m=1:length(obj.dm)
                    if ((dJ(m) - dJ(mu(1)))>0 && obj.dm(m)==0)
                        D(m) = 0;
                    elseif (m==mu(1))
                            D(m) = sum(dJ) - (length(dm)+1)*dJ(mu(1));
                    else
                        D(m) = -dJ(m)+dJ(mu(1));
                    end
                end
                Jcross = 0;
                dcross = obj.dm;
                Dcross = D;
                while (Jcross<J)
                    obj.dm = dcross;
                    D = Dcross;
                    [~, nu] = min(-obj.dm(D<0)./D(D<0));
                    gamma = -obj.dm(nu)/D(nu);
                    dcross = obj.dm+gamma*D;
                    Dcross(mu) = D(mu)-D(nu);
                    Dcross(nu) = 0;
                    K = obj.SD.computeKernelCombination(dcross,1:obj.SD.n);
                    SM = simpleModelCSVM(K,obj.C,obj.Il, obj.Iu, obj.Iw);
                    SM.train;
                    Jcross = (SM.alpha.*obj.SD.trainlab([obj.Iw,obj.Iu]))*K([SM.Iw;SM.Iu],[SM.Iw,SM.Iu])*...
                    (SM.alpha.*obj.SD.trainlab([obj.Iw,obj.Iu]))' - sum((SM.alpha.*obj.SD.trainlab([obj.Iw,obj.Iu])));
                end
                % optimiser gamma!!
                obj.dm = obj.dm+(gamma/2)*Dcross;
                
            end
            
        end
        
        
        
        
        function solveLinearSystem(obj)
            be = -1;
            indd = obj.SD.getKernelValues(obj.Iw,obj.Iu);
            ce = diag(obj.SD.kernelCache(obj.Iw,indd))';
            if ~isempty(obj.Iu)
                ce = ce + 1*(obj.Slacks(obj.Iu))'*obj.SD.kernelCache(obj.Iu,indd)
                be = be + 1*obj.Slacks(obj.Iu)'*obj.SD.trainlab(obj.Iu)
            end
            if obj.optimType==1
                G = chol(obj.SD.kernelCache(obj.Iw,indd)+obj.kktgap*eye(length(obj.Iw)));
                M = obj.SD.trainlab(obj.Iw)'*(G\(G'\obj.SD.trainlab(obj.Iw)));
                obj.b = M\(obj.SD.trainlab(obj.Iw)'*(G\(G'\ce'))+be);
                obj.alpha = ( G\(G'\(ce'-obj.SD.trainlab(obj.Iw)*obj.b)));
            else
                error('not implemented yet');
            end
        end
        
        
        function reinit(obj)
            obj.Iw = [find(obj.SD.trainlab==-1,1),find(obj.SD.trainlab==1,1)];
            obj.Iu = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            
            if length(obj.Slacks)~=length(obj.SD.trainlab)
                if length(obj.C)==length(obj.SD.trainlab)
                    obj.Slacks = obj.C;
                elseif length(obj.C)==length(obj.labs)
                    for i=1:length(obj.labs)
                        obj.Slacks(obj.SD.trainlab==obj.labs(i)) = obj.C(i);
                    end
                else
                    obj.Slacks = obj.C(1)*ones(size(obj.SD.trainlab));
                    
                end
            end
            obj.activated = 0;
            obj.converged = 0;
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelSVDD:print2D','must be 2D data');
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
            %f = kernelEval(obj,xtest);
            f = kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ...
                    obj.SD.trainlab([obj.Iw,obj.Iu]), xtest, ones(size(xtest,2),1))';
            if obj.converged==0
                ypred = ([obj.alpha;obj.Slacks(obj.Iu)].*obj.SD.trainlab([obj.Iw,obj.Iu]))'*f+obj.b;
            else
                ypred = obj.alpha'*f+obj.b;
            end
            ypredmat=reshape(ypred,na,nb);
            
            hold off;
            plot(obj.SD.trainvec(1,:),obj.SD.trainvec(2,:),'k+');
            hold on;
            plot(obj.SV(1,:),obj.SV(2,:),'om','MarkerSize',10,'LineWidth',2);        
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ 0 0],'g');
            set(h,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[ -1 1],'y');
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        function [y] = evaluateSolution(obj,candidates)
            indTR = obj.SD.getKernelValues(candidates,candidates);
            TR = diag(obj.SD.kernelCache(candidates,indTR));
            indd = obj.SD.getKernelValues([obj.Iw,obj.Iu],candidates);
            [val,inds]= sort(indd);
            temp = [obj.alpha;obj.Slacks(obj.Iu)];
            a = obj.SD.kernelCache(candidates,val)*temp(inds);
            R = obj.SD.kernelCache(obj.Iw(1),indd(1)) - 2*obj.SD.kernelCache(obj.Iw(1),val)*temp(inds);
            y = R - TR + 2*a;
        end
        
        
        function [ypred perf] = test(obj,testdata,testlabel)
            
            indd = obj.SD.getKernelValues(obj.Iw(1),obj.Iw(1));
            %[val,inds]= sort(indd);
            temp = [obj.alpha;obj.Slacks(obj.Iu)];
            %a = obj.SD.kernelCache(candidates,val)*temp(inds);
            R = obj.SD.kernelCache(obj.Iw(1),indd) - 2*obj.SD.kernelCache(obj.Iw(1),[obj.Iw,obj.Iu])*temp;
            %y = R - diag(obj.SD.kernelCache(candidates,val)) + 2*a;
            
            if strcmp(obj.SD.kernelType,'precomputed') % pre-computed case
                assert(size(testdata,1)==obj.SD.n,'simpleModelSVDD:wrongInput','precomputed-kernel used, need to provide a matrix nntrain*nbtest to compute a test error');
                ypred = R - diag(testdata) + 2*(obj.alpha'*testdata([obj.Iw,obj.Iu],:))';
                if nargin==3
                    perf = 100*length(find(sign(ypred)==testlabel))/length(testlabel);
                else
                    perf = -1;
                end
            else
                %ypred = (obj.alpha'*kernelEval(obj,testdata)+obj.b)';
                K = kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ...
                    abs(obj.SD.trainlab([obj.Iw,obj.Iu])), testdata, ones(size(testdata,2),1));
                K2 =  kernelTest(obj.SD.kernelType,obj.SD.kernelParam,testdata, ...
                     ones(size(testdata,2),1), testdata, ones(size(testdata,2),1));
                ypred = R - diag(K2) + 2*(K*obj.alpha);
                if nargin==3
                    perf = 100*length(find(sign(ypred)==testlabel))/length(testlabel);
                else
                    perf = -1;
                end
            end
        end
        
    end
    
end
