classdef simpleModelNPSVR < simpleModel
    % SIMPLEMODELCSVM This class inherits from simpleModel. It contains
    % particular methods to solve binary-C-SVM.
    % This class support C-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    %  9/01/2014
    % Last Modified 9/01/2014 by G.L.
    properties
        etube; % epsilon tube
        targets;
        negs;
        V;
    end
    
    
    methods
        % contructor
        function SM = simpleModelNPSVR(SD,C,etube,nbEig,Il,Iu,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelCSVM object
            
            assert(nargin==4 || nargin==7,'simpleModelNPSVR:wrongInput','Requires 3 or 6 input arguments');
            if nargin<2
                error('not enough input');
            else
                
                if (strcmp(SD.kernelType,'precomputed'))
                    SM.SD = simpleData([SD.trainvec, -SD.trainvec;-SD.trainvec,SD.trainvec],[ones(size(SD.trainlab));-ones(size(SD.trainlab))]);
                    SM.SD.kernelCache = SM.SD.trainvec;
                    SM.SD.kernelType = 'precomputed';
                else
                    SM.SD = simpleData([SD.trainvec, SD.trainvec],[ones(size(SD.trainlab));-ones(size(SD.trainlab))]);
                    SM.SD.getKernelValues(1:SD.n);
                end
                SM.targets = [SD.trainlab-etube; -SD.trainlab-etube];
                [SM.V D] = eigs(SM.SD.kernelCache(1:SM.SD.n,1:SM.SD.n),min(nbEig,SM.SD.n-2));
                SM.negs = find(diag(D)<0);
                SM.SD.kernelCache(1:SM.SD.n,1:SM.SD.n) = SM.V*abs(D)*SM.V';
                SM.C = C;
                SM.Slacks = zeros(SD.n,1);
                if nargin==7
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [1,2];
                    SM.Iu = [];
                    SM.Il = 1:SM.SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(length(SM.Iw),1);
                end
                SM.labs = sort(unique(SM.SD.trainlab));
                if length(C)==length(SM.SD.trainlab)
                    SM.Slacks = C;
                elseif length(C)==length(SM.labs)
                    for i=1:length(SM.labs)
                        SM.Slacks(SM.SD.trainlab==SM.labs(i)) = C(i);
                    end
                else
                    SM.Slacks = C(1)*ones(size(SM.SD.trainlab));
                    
                end
                SM.activated = 0;
                SM.converged = 0;
                SM.etube = etube;
            end

        end
        
        
        
        function set.negs(obj,val)
            obj.negs = val;
        end
        
        function val = get.negs(obj)
            val = obj.negs;
        end
        
        function set.V(obj,val)
            obj.V = val;
        end
        
        function val = get.V(obj)
            val = obj.V;
        end
        
        function val = get.etube(obj)
            val = obj.etube;
        end
        
        
        function set.etube(obj,val)
            obj.etube = val;
        end
        
        function val = get.targets(obj)
            val = obj.targets;
        end
        
        
        function set.targets(obj,val)
            obj.targets = val;
        end
        
        
        function solveLinearSystem(obj)
            ce = obj.targets(obj.Iw)';
            be = 0;
            indd = obj.SD.getKernelValues(obj.Iw,obj.Iu);
            if ~isempty(obj.Iu)
                ce = ce - (obj.Slacks(obj.Iu))'*obj.SD.kernelCache(obj.Iu,indd);
                be = be - obj.Slacks(obj.Iu)'*obj.SD.trainlab(obj.Iu);
            end
            if obj.optimType==1
                G = chol(obj.SD.kernelCache(obj.Iw,indd)+obj.kktgap*eye(length(obj.Iw)));
                M = obj.SD.trainlab(obj.Iw)'*(G\(G'\obj.SD.trainlab(obj.Iw)));
                obj.b = M\(obj.SD.trainlab(obj.Iw)'*(G\(G'\ce'))-be);
                obj.alpha = G\(G'\(ce'-obj.SD.trainlab(obj.Iw)*obj.b));
            else
                error('not implemented yet');
            end
        end
        
        function y = evaluateSolution(obj,candidates)
            indd = obj.SD.getKernelValues([obj.Iw,obj.Iu],candidates);
            [val,inds]= sort(indd);
            temp = [obj.alpha;obj.Slacks(obj.Iu)];
            y = obj.SD.kernelCache(candidates,val)*temp(inds) + obj.b*obj.SD.trainlab(candidates) - obj.targets(candidates);
        end
        
        function reinit(obj)
            obj.Iw = [1,2];
            obj.Iu = [];
            obj.Il = 1:obj.SD.n/2;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            
            if length(obj.Slacks)~=length(obj.SD.trainlab)
                if length(obj.C)==length(obj.SD.trainlab)
                    obj.Slacks = obj.C;
                elseif length(obj.C)==length(obj.labs)
                    for i=1:length(obj.labs)
                        obj.Slacks(obj.SD.trainlab==obj.labs(i)) = obj.C(i);
                    end
                else
                    obj.Slacks = obj.C(1)*ones(size(obj.SD.trainlab));
                    
                end
            end
            obj.activated = 0;
            obj.converged = 0;
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==1,'simpleModelCSVM:printD','must be 1D data');
            xtest=linspace(min(obj.SD.trainvec),max(obj.SD.trainvec),1000);
            obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
            %f = kernelEval(obj,xtest);
            f = kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ...
                abs(obj.SD.trainlab([obj.Iw,obj.Iu])), xtest, ones(size(xtest,2),1))';
            if obj.converged==0
                ypred = ([obj.alpha;obj.Slacks(obj.Iu)].*obj.SD.trainlab([obj.Iw,obj.Iu]))'*f+obj.b;
            else
                ypred = obj.alpha'*f+obj.b;
            end
            
            hold off;
            plot(obj.SD.trainvec(1:obj.SD.n/2),obj.targets(1:obj.SD.n/2)+obj.etube,'-k');
            hold on;
            plot(obj.SD.trainvec(1:obj.SD.n/2),obj.targets(1:obj.SD.n/2),'-.k');
            plot(obj.SD.trainvec(obj.SD.n/2+1:end),-obj.targets(obj.SD.n/2+1:end),'-.k');
            plot(obj.SV,obj.targets([obj.Iw,obj.Iu]).*obj.SD.trainlab([obj.Iw,obj.Iu]),'om','MarkerSize',10,'LineWidth',2);
            plot(xtest,ypred,'-rx');
            
            drawnow;
            
        end
        
        function [ypred perf] = test(obj,testdata,testlabel)
            if strcmp(obj.SD.kernelType,'precomputed') % pre-computed case
                assert(size(testdata,1)==obj.SD.n,'simpleModel:wrongInput','precomputed-kernel used, need to provide a matrix nntrain*nbtest to compute a test error');
                ypred = (obj.alpha'*testdata([obj.Iw,obj.Iu],:)+obj.b)';
                if nargin==3
                    perf = mean(abs(testlabel-ypred));
                else
                    perf = -1;
                end
            else
                %ypred = (obj.alpha'*kernelEval(obj,testdata)+obj.b)';
                ypred = (obj.alpha'*kernelTest(obj.SD.kernelType,obj.SD.kernelParam,obj.SD.trainvec(:,[obj.Iw,obj.Iu]), ...
                    abs(obj.SD.trainlab([obj.Iw,obj.Iu])), testdata, ones(size(testdata,2),1))'+obj.b)';
                if nargin==3
                    perf = corr(testlabel,ypred);
                else
                    perf = -1;
                end
            end
        end
        
        
        function finish(obj)
            obj.alpha = [obj.alpha;obj.Slacks(obj.Iu)];
            obj.alpha = obj.V([obj.Iw,obj.Iu],:)'*obj.alpha;
            obj.alpha(obj.negs) = -obj.alpha(obj.negs);
            obj.alpha = (obj.V*obj.alpha).*obj.SD.trainlab;
            if not(strcmp(obj.SD.kernelType,'precomputed'))
                obj.SV = obj.SD.trainvec;
                obj.Iw= 1:obj.SD.n;
                obj.Iu = [];
            else
                obj.Iw= 1:obj.SD.n;
                obj.Iu = [];
            end
        end
        
        
        function print2DMulti(obj,SMt)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            assert(nargin==2);
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            if length(SMt)==length(unique(obj.SD.trainlab));
                ypred = obj.test1vsAll(SMt,xtest);
            else
                ypred = obj.test1vs1(SMt,xtest);
            end
            ypredmat=reshape(ypred,na,nb);
            %labs = unique(obj.SD.trainlab);
            hold on;
            [c,h] = contour(xtesta1,xtesta2,ypredmat);
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        
        function [err,p] = crossValError(obj,nbFold)
            assert(nbFold<=obj.SD.n && nbFold>0,'simpleModel:wrongIput','nbFold must range between 1 and number of training points');
            Sl = obj.Slacks;
            SDSaved = obj.SD;
            targetsSaved = obj.targets;
                s = ceil(obj.SD.n/(2*nbFold));
                p = zeros(1,nbFold);
                rn = randperm(SDSaved.n/2);
            if ~isempty(obj.SD.trainvec)
                for i=1:nbFold
                    il = rn((i-1)*s+1:min(i*s,SDSaved.n/2));
                    itt = setdiff(1:SDSaved.n/2,il);
                    it = [itt,itt+SDSaved.n/2];
                    
                    if strcmp(SDSaved.kernelType,'precomputed')
                        SDt = simpleData([SDSaved.trainvec(it,it)],[SDSaved.trainlab(it)]);
                        SDt.kernelCache = SDt.trainvec;
                        SDt.kernelType = 'precomputed';
                    else
                        %SDt = simpleData(SDSaved.trainvec(:,it),SDSaved.trainlab(it));
                    end
                    SDt.kernelType = SDSaved.kernelType;
                    SDt.kernelParam = SDSaved.kernelParam;
                    obj.SD = SDt;
                    obj.targets = targetsSaved(it);
                    obj.clear();
                    obj.reinit();
                    obj.Slacks = Sl(it);
                    [obj.V D] = eigs(obj.SD.kernelCache(1:obj.SD.n,1:obj.SD.n),floor((nbFold-1)*size(obj.V,2)/(nbFold)));
                    obj.negs = find(diag(D)<0);
                    obj.SD.kernelCache(1:obj.SD.n,1:obj.SD.n) = obj.V*abs(D)*obj.V';
                    obj.train();
                    %obj.finish();
                    if strcmp(SDSaved.kernelType,'precomputed')
                        [~ , p(i)] = obj.test([SDSaved.trainvec(it,il)],targetsSaved(il)+obj.etube);
                    else
                        [~ , p(i)] = obj.test(SDSaved.trainvec(:,il),SDSaved.targets(il)+obj.etube);
                    end
                   clear SDt;
                end
            else
                SDSaved.kernelCache = SDSaved.kernelCache.*(SDSaved.trainlab*SDSaved.trainlab');
                
                for i=1:nbFold
                    il = rn((i-1)*s+1:min(i*s,SDSaved.n));
                    it = setdiff(1:SDSaved.n,il);
                    SDt = simpleData(SDSaved.kernelCache(it,it),SDSaved.trainlab(it));
                    obj.SD = SDt;
                    obj.Slacks = Sl(it);
                    obj.clear();
                    obj.reinit();
                    obj.train();
                    [~ , p(i)] = obj.test(SDSaved.kernelCache(it,il),SDSaved.trainlab(il));
                    clear SDt;
                end
                SDSaved.kernelCache = SDSaved.kernelCache.*(SDSaved.trainlab*SDSaved.trainlab');
            end
            err = mean(p);
            obj.SD = SDSaved();
            obj.Slacks = Sl;
            obj.targets = targetsSaved;
        end
        
    end
    
end
