classdef simpleModelNuSVM < simpleModel
    % SIMPLEMODELCSVM This class inherits from simpleModel. It contains
    % particular methods to solve binary-Nu-SVM.
    % This class support C-SVM solver, warm-start, 2D solution plot, and
    % cross-validation error-rate evaluation (see simpleModel.crossValError).
    % See also : simpleModel, simpleData
    
    % Create by Galle Loosli
    % 18/02/2011
    % Last Modified 18/02/2011 by G.L.
    
    properties
       rho; 
        
    end
    
    methods
        % contructor
        function SM = simpleModelNuSVM(SD,C,Il,Iu,Iw)
            % simpleModelCSVM : constructor.
            % Input :   SD : a simpleData object
            %           C : Slack value (can be of length 1 for all points
            %           equally penalized, or one value by class, or one
            %           value per training point)
            %           Il : indices list of non support vectors for warm-start
            %           Iu : indices list of upper bounded support vectors for
            %           warm-start
            %           Iw : indices list of non-bounded support vectors
            % Number of input arguments must be at least 2 (simpleData
            % object and simpleProblem object) or exactly 5 (for warm-start,
            % one can provide the initial repartition of support vectors)
            % Output :   SM : a simpleModelCSVM object
            
            assert(nargin==2 || nargin==5,'simpleModelNuSVM:wrongInput','Requires 2 or 5 input arguments');
            if nargin<2
                error('not enough input');
            else
                SM.SD = SD;
                assert(length(C)==1,'simpleModelNuSVM:wrongInput','Only one value for nu');
                SM.C = C*SD.n;
                SM.Slacks = ones(1,SD.n);
                if nargin==5
                    SM.Il = Il;
                    SM.Iu = Iu;
                    SM.Iw = Iw;
                    SM.oldalpha = zeros(length(Iw),1);
                else
                    SM.Iw = [find(SD.trainlab==-1,1),find(SD.trainlab==1,1)];
                    SM.Iu = [];
                    SM.Il = 1:SD.n;
                    SM.Il = setdiff(SM.Il,SM.Iw);
                    SM.oldalpha = zeros(2,1);
                end
                SM.labs = sort(unique(SD.trainlab));
                SM.activated = 0;
                SM.converged = 0;
                SM.b = 0;
            end
        end
        
        
        function set.rho(obj,val)
            obj.rho = val;
        end
        
        function rho = get.rho(obj)
            rho = obj.rho;
        end
        
        
        function activateBoxConstraints(obj)
            if (~isempty(find(obj.alpha<0, 1)) || ~isempty(find(obj.alpha>1, 1))) && sum(obj.alpha)+length(obj.Iu)>=obj.C
                dir = obj.alpha - obj.oldalpha;
                indad = find(obj.alpha < 0);
                indsup = find(obj.alpha > 1);
                [tI indmin] = min(-obj.oldalpha(indad)./dir(indad));
                [tS indS] = min((1-obj.oldalpha(indsup))./dir(indsup));
                if isempty(tI) , tI = tS + 1; end;
                if isempty(tS) , tS = tI + 1; end;
                t = min(tI,tS);
                obj.oldalpha = obj.oldalpha + t*dir;
                if t==tI
                    obj.Il = sort([obj.Il , obj.Iw(indad(indmin))]);
                    obj.Iw(indad(indmin)) = [];
                    obj.oldalpha(indad(indmin)) = [];
                    obj.alpha(indad(indmin)) = [];
                else
                    obj.Iu = sort([obj.Iu , obj.Iw(indsup(indS))]);
                    obj.Iw(indsup(indS))= [];
                    obj.oldalpha(indsup(indS))= [];
                    obj.alpha(indsup(indS)) = [];
                end
                obj.activated = 1;
            else
                obj.activated = 0;
            end
            
        end
        
        
        function y = evaluateSolution(obj,candidates)
            indd = obj.SD.getKernelValues([obj.Iw,obj.Iu],candidates);
            [val,inds]= sort(indd);
            temp = [obj.alpha;obj.Slacks(obj.Iu)];
            y = obj.SD.kernelCache(candidates,val)*temp(inds)-obj.SD.trainlab(candidates)*obj.rho;
        end
        
        
        function relaxBoxConstraints(obj)
            xt = zeros(obj.SD.n,1);
            xt(obj.Iw) = obj.alpha;
            xt(obj.Iu) = 1;
            if not(isempty(obj.Il))
                nul = obj.evaluateSolution(obj.Il)-obj.SD.trainlab(obj.Il)*obj.rho;
                [mm mpos]=min(nul);
            else
                mm = inf; mpos = [];
            end
            if not(isempty(obj.Iu))
                nuu = obj.evaluateSolution(obj.Iu)-obj.SD.trainlab(obj.Iu)*obj.rho;
                [mmS mposS]=min(-nuu);
            else
                mmS = inf; mposS = [];
            end
            
            if (~isempty(mmS) && mmS < -obj.kktgap) || (~isempty(mm) && mm < -obj.kktgap)
                if isempty(mmS) || mm < mmS
                    obj.Iw = sort([obj.Iw , obj.Il(mpos)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.Il(mpos) = [];
                else
                    obj.Iw = sort([obj.Iw , obj.Iu(mposS)]);
                    obj.oldalpha = xt(obj.Iw);
                    obj.Iu(mposS)=[];
                end
            else
                obj.converged = 1;
            end
        end

        
        function solveLinearSystem(obj)
            
            ce = zeros(1,length(obj.Iw));
            %be = 0;
            be = obj.C;
            indd = obj.SD.getKernelValues(obj.Iw,obj.Iu);
            if ~isempty(obj.Iu)
                ce = ce - ones(1,length(obj.Iu))*obj.SD.kernelCache(obj.Iu,indd);
                %be = be - sum(obj.SD.trainlab(obj.Iu));
                be = be - length(obj.Iu);
            end
            if obj.optimType==1
                G = chol(obj.SD.kernelCache(obj.Iw,indd)+obj.kktgap*eye(length(obj.Iw)));
                M = ones(1,length(obj.Iw))*(G\(G'\ones(length(obj.Iw),1)));
                obj.rho = (M*obj.SD.n)\(ones(1,length(obj.Iw))*(G\(G'\ce'))+be);
                obj.alpha = G\(G'\(ce'+ones(length(obj.Iw),1)*obj.rho));
            else
                error('not implemented yet');
            end
        end
        
        function finish(obj)
            if ~isempty(obj.SD.trainvec)
                obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
                
            else
                %disp('no training data can be stored for later use :
                %pre-computed matrix');
            end
            obj.alpha = [obj.alpha;ones(length(obj.Iu),1)].*obj.SD.trainlab([obj.Iw,obj.Iu]);
        end
        
        
        function reinit(obj)
            obj.Iw = [find(obj.SD.trainlab==-1,1),find(obj.SD.trainlab==1,1)];
            obj.Iu = [];
            obj.Il = 1:obj.SD.n;
            obj.Il = setdiff(obj.Il,obj.Iw);
            obj.oldalpha = zeros(2,1);
            obj.activated = 0;
            obj.converged = 0;
            obj.b = 0;
        end
        
        function print2D(obj)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            obj.SV = obj.SD.trainvec(:,[obj.Iw,obj.Iu]);
            f = kernelEval(obj,xtest);
            if obj.converged==0
                ypred = ([obj.alpha;ones(length(obj.Iu),1)].*obj.SD.trainlab([obj.Iw,obj.Iu]))'*f;
            else
                ypred = obj.alpha'*f;
            end
            ypredmat=reshape(ypred,na,nb);
            
            hold off;
            plot(obj.SD.trainvec(1,:),obj.SD.trainvec(2,:),'k+');
            hold on;
            plot(obj.SV(1,:),obj.SV(2,:),'om','MarkerSize',10,'LineWidth',2);
            [c,h] = contour(xtesta1,xtesta2,ypredmat,[-obj.rho 0 obj.rho],'g');
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        
        
        function print2DMulti(obj,SMt)
            assert(size(obj.SD.trainvec,1)==2,'simpleModelCSVM:print2D','must be 2D data');
            assert(nargin==2);
            scale = 2;
            [xtesta1,xtesta2]=meshgrid([-scale:0.1:scale],[-scale:0.1:scale]);
            [na,nb]=size(xtesta1);
            xtest1=reshape(xtesta1,1,na*nb);
            xtest2=reshape(xtesta2,1,na*nb);
            xtest=[xtest1;xtest2];
            if length(SMt)==length(unique(obj.SD.trainlab));
                ypred = obj.test1vsAll(SMt,xtest);
            else
                ypred = obj.test1vs1(SMt,xtest);
            end
            ypredmat=reshape(ypred,na,nb);
            labs = unique(obj.SD.trainlab);
            hold on;
            [c,h] = contour(xtesta1,xtesta2,ypredmat);
            set(h,'LineWidth',2);
            drawnow;
            
        end
        
        function SMt = train1vs1(obj,SMt)
            assert(isa(obj,'simpleModel'));
            nbClass = length(obj.labs);
            k = 0;
            if nargin==1
                for i=1:nbClass
                    for j=i+1:nbClass
                        k=k+1;
                        ind = sort([find(obj.SD.trainlab==obj.labs(i));find(obj.SD.trainlab==obj.labs(j))]);
                        tl = obj.SD.trainlab(ind);
                        tl(tl==obj.labs(i))=-1;
                        tl(tl==obj.labs(j))=1;
                        if isempty(obj.SD.trainvec)
                            tt = obj.SD.kernelCache(ind,ind);
                        else
                            tt = obj.SD.trainvec(:,ind);
                        end
                        SDt{k} = simpleData(tt,tl);
                        SMt{k} = feval(class(obj),SDt{k},obj.C);
                        SMt{k}.train();
                    end
                end
            elseif nargin==2
                assert(length(SMt)==(nbClass-1)*nbClass/2);
                for i=1:nbClass
                    for j=i+1:nbClass
                        k=k+1;
                        if (SMt{k}.SD.kernelParam ~= obj.SD.kernelParam || SMt{k}.SD.kernelType ~= obj.SD.kernelType)
                            SMt{k}.SD.clear;
                            SMt{k}.SD.kernelType = obj.SD.kernelType;
                            SMt{k}.SD.kernelParam = obj.SD.kernelParam;
                        end
                        
                        SMt{k} = feval(class(obj),SMt{k}.SD,obj.C,SMt{k}.Il,SMt{k}.Iu,SMt{k}.Iw);
                        SMt{k}.train();
                    end
                end
            end
        end
        
        
        
        function SMt = train1vsAll(obj,SMt)
            assert(isa(obj,'simpleModel'));
            nbClass = length(obj.labs);
            if nargin==1
                for i=1:nbClass
                    tl = obj.SD.trainlab;
                    tl(tl~=obj.labs(i))=-1;
                    tl(tl==obj.labs(i))=1;
                    if isempty(obj.SD.trainvec)
                        SDt{i} = simpleData(obj.SD.kernelCache,tl);
                    else
                        SDt{i} = simpleData(obj.SD.trainvec,tl);
                    end
                    SMt{i} = feval(class(obj),SDt{i},obj.C);
                    SMt{i}.train();
                end
            elseif nargin==2
                assert(length(SMt)==nbClass);
                for i=1:nbClass
                    if (SMt{i}.SD.kernelParam ~= obj.SD.kernelParam || SMt{i}.SD.kernelType ~= obj.SD.kernelType)
                        SMt{i}.SD.clear;
                        SMt{i}.SD.kernelType = obj.SD.kernelType;
                        SMt{i}.SD.kernelParam = obj.SD.kernelParam;
                    end
                    SMt{i} = feval(class(obj),SMt{i}.SD,obj.C,SMt{i}.Il,SMt{i}.Iu,SMt{i}.Iw);
                    SMt{i}.train();
                end
            end
        end
        
        
    end
    
end
